import logging
import random

from trosnoth.const import (
    NOT_ENOUGH_STARS_REASON, HAVE_UPGRADE_REASON, PLAYER_DEAD_REASON,
    GAME_NOT_STARTED_REASON, ALREADY_TURRET_REASON, TOO_CLOSE_TO_EDGE_REASON,
    TOO_CLOSE_TO_ORB_REASON, NOT_IN_DARK_ZONE_REASON, ALREADY_DISRUPTED_REASON,
    INVALID_UPGRADE_REASON, DISABLED_UPGRADE_REASON,
)
from trosnoth.messages.base import (
    AgentRequest, ServerCommand, ServerResponse, ClientCommand,
)

log = logging.getLogger(__name__)


EDGE_TURRET_BOUNDARY = 100
ORB_TURRET_BOUNDARY = 150


class AbandonUpgradeMsg(ClientCommand):
    idString = 'DUpg'
    fields = 'playerId', 'tickId'
    packspec = 'cH'
    timestampedPlayerRequest = True

    def clientValidate(self, localState, world, sendResponse):
        player = localState.player
        if player.upgrade is not None and player.upgrade.upgradeType != 'g':
            return True
        return False

    def applyRequestToLocalState(self, localState):
        localState.player.deleteUpgrade()

    def serverApply(self, game, agent):
        player = game.world.getPlayer(self.playerId)
        if player.upgrade is not None and player.upgrade.upgradeType != 'g':
            game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        player.deleteUpgrade(abandoned=True)


class BuyUpgradeMsg(AgentRequest):
    '''
    Signal from interface that a buy has been requested.

    In order for everything to stay in sync, an upgrade purchase works in 3
    steps.

    1. The client sends BuyUpgradeMsg.
    2. The server checks that the stars are avaliable, and replies with
        UpgradeApprovedMsg.
    3. The client sends PlayerHasUpgradeMsg, which is then broadcast to all
        clients.

    To make launching of grenades more responsive, there is a special flag on
    Upgrade classes called doNotWaitForServer. If this flag is set, the client
    assumes that the purchase is successful unless it hears otherwise, and the
    server broadcasts PlayerHasUpgradeMsg immediately at step 2. Step 3 is not
    performed in this case.
    '''
    idString = 'GetU'
    fields = 'upgradeType', 'tickId'
    packspec = 'cH'
    timestampedPlayerRequest = True

    def clientValidate(self, localState, world, sendResponse):
        upgrade = world.getUpgradeType(self.upgradeType)
        if not upgrade.doNotWaitForServer:
            # Don't bother trying to validate locally, let the server decide.
            return True

        player = localState.player
        denialReason = self.getDenialReason(player, upgrade)
        if denialReason:
            response = CannotBuyUpgradeMsg(denialReason)
            response.local = True
            sendResponse(response)
            return False
        return True

    def applyRequestToLocalState(self, localState):
        upgrade = localState.world.getUpgradeType(self.upgradeType)
        if upgrade.doNotWaitForServer:
            localState.player.upgradeUsed(self.upgradeType, local=localState)

    def serverApply(self, game, agent):
        player = agent.player
        if player is None:
            return

        upgrade = player.world.getUpgradeType(self.upgradeType)
        denialReason = self.getDenialReason(player, upgrade)
        if denialReason:
            agent.messageToAgent(CannotBuyUpgradeMsg(denialReason))
        else:
            self._processUpgradePurchase(game, agent, player, upgrade)

    def getDenialReason(self, player, upgrade):
        from trosnoth.model.upgrades import allUpgrades
        if player.upgrade is not None:
            return HAVE_UPGRADE_REASON
        if player.pendingUpgrade is not None:
            return HAVE_UPGRADE_REASON
        if not player.world.abilities.upgrades:
            return GAME_NOT_STARTED_REASON
        if player.dead:
            return PLAYER_DEAD_REASON

        if upgrade not in allUpgrades:
            return INVALID_UPGRADE_REASON

        if not upgrade.enabled:
            return DISABLED_UPGRADE_REASON

        if player.getTeamStars() < upgrade.requiredStars:
            return NOT_ENOUGH_STARS_REASON

        return self._checkUpgradeConditions(player, upgrade)

    def _checkUpgradeConditions(self, player, upgrade):
        '''
        Checks whether the conditions are satisfied for the given player to be
        able to purchase an upgrade of the given kind. Returns None if no
        condition is violated, otherwise returns a one-byte reason code for why
        the upgrade cannot be purchased.
        '''
        from trosnoth.model.upgrades import Turret, MinimapDisruption

        if upgrade is Turret:
            zone = player.getZone()
            if not zone:
                return NOT_IN_DARK_ZONE_REASON
            if not (player.isFriendsWithTeam(zone.owner) and zone.dark):
                return NOT_IN_DARK_ZONE_REASON
            if zone.turretedPlayer is not None:
                return ALREADY_TURRET_REASON
            if player.getMapBlock().fromEdge(player) < EDGE_TURRET_BOUNDARY:
                return TOO_CLOSE_TO_EDGE_REASON

            distanceFromOrb = player.distanceFromCentre()
            if distanceFromOrb < ORB_TURRET_BOUNDARY:
                return TOO_CLOSE_TO_ORB_REASON
        elif upgrade is MinimapDisruption:
            if player.team is None or player.team.usingMinimapDisruption:
                return ALREADY_DISRUPTED_REASON

        return None

    def _processUpgradePurchase(self, game, agent, player, upgrade):
        '''
        Sends the required sequence of messages to gameRequests to indicate
        that the upgrade has been purchased by the player.
        '''
        remaining = upgrade.requiredStars

        # Take from the purchasing player first.
        fromPurchaser = min(player.stars, remaining)
        game.sendServerCommand(PlayerStarsSpentMsg(player.id, fromPurchaser))
        remaining -= fromPurchaser

        # Now take evenly from other team members.
        if remaining > 0:
            players = []
            totalStars = 0
            for p in game.world.players:
                if p.team == player.team and p.id != player.id and p.stars > 0:
                    players.append(p)
                    totalStars += p.stars
            # Order by descending number of stars
            # Shuffle first to avoid biasing any player with the same number of
            # stars as another
            random.shuffle(players)
            players.sort(cmp=lambda p1, p2: cmp(p1.stars, p2.stars))
            players.reverse()

            for p in players:
                fraction = remaining / (totalStars + 0.)
                toGive = int(round(fraction * p.stars))
                remaining -= toGive
                totalStars -= p.stars
                game.sendServerCommand(PlayerStarsSpentMsg(p.id, toGive))
            # Everything should add up
            assert remaining == 0
            assert totalStars == 0

        if upgrade.doNotWaitForServer:
            game.sendServerCommand(PlayerHasUpgradeMsg(
                upgrade.upgradeType, self.tickId, player.id))
        else:
            player.pendingUpgrade = upgrade.upgradeType
            agent.messageToAgent(UpgradeApprovedMsg(upgrade.upgradeType))
        player.onStarsSpent(upgrade.requiredStars)


class PlayerHasUpgradeMsg(ClientCommand):
    '''
    Sent by the local client when it receives word from the server that an
    upgrade purchase has been approved.
    '''
    idString = 'GotU'
    fields = 'upgradeType', 'tickId', 'playerId'
    packspec = 'cHc'
    playerId = '\x00'
    timestampedPlayerRequest = True

    def applyRequestToLocalState(self, localState):
        localState.player.upgradeUsed(self.upgradeType, local=localState)
        # Guaranteed to be verified if the client is sending this in response
        # to UpgradeApprovedMsg.
        localState.player.upgrade.verified = True

    def serverApply(self, game, agent):
        if agent.player and agent.player.pendingUpgrade == self.upgradeType:
            self.playerId = agent.player.id
            agent.player.pendingUpgrade = None
            game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        player.upgradeUsed(self.upgradeType)

    def applyOrderToLocalState(self, localState, world):
        if localState.player and self.playerId == localState.player.id:
            upgrade = localState.player.upgrade
            if upgrade:
                upgrade.serverVerified(localState)


class PlayerStarsSpentMsg(ServerCommand):
    '''
    This message is necessary, because without it a client can't always tell
    where the PlayerHasUpgradeMsg pulled all its stars from.
    '''
    idString = 'Spnt'
    fields = 'playerId', 'count'
    packspec = 'cI'


class AwardPlayerStarMsg(ServerCommand):
    idString = 'Star'
    fields = 'playerId'
    packspec = 'c'

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        player.incrementStars()


class CannotBuyUpgradeMsg(ServerResponse):
    '''
    Tag is the value sent in BuyUpgradeMsg.
    Valid reasonId values are defined in trosnoth.const.
    '''
    idString = 'NotU'
    fields = 'reasonId'
    packspec = 'c'
    local = False

    def applyOrderToLocalState(self, localState, world):
        if self.local:
            return
        if localState.player:
            upgrade = localState.player.upgrade
            if upgrade and not upgrade.verified:
                upgrade.deniedByServer(localState)


class UpgradeApprovedMsg(ServerResponse):
    '''
    Signals to the player that's trying to buy an upgrade that the purchase has
    been successful, and the player should proceed to send a
    PlayerHasUpgradeMsg to use the upgrade. This back and forth is necessary
    because the using of the upgrade needs to happen based on the location of
    the player on their own client's screen, not based on the server's idea of
    where the client is.
    '''
    idString = 'ByOk'
    fields = 'upgradeType'
    packspec = 'c'


class UpgradeChangedMsg(ServerCommand):
    '''
    A message for the clients that informs them of a change in an upgrade stat.
    statType may be:
        'S' - star cost
        'T' - time limit
        'X' - explosion radius
        'E' - enabled
    '''
    idString = 'UpCh'
    fields = 'upgradeType', 'statType', 'newValue'
    packspec = 'ccf'



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import os

user = object() # Used to write/read user data.

USERROOT = os.path.expanduser('~')
USERHOME = os.path.join(USERROOT, '.trosnoth')

def getPath(module, *bits):
    if module is user:
        makeDirs(USERHOME)
        return os.path.join(USERHOME, *bits)
    if _datapath is None:
        return os.path.join(os.path.dirname(module.__file__), *bits)
    return os.path.join(_datapath, module.__name__.replace('.','/'), *bits)

def makeDirs(path):
    if not os.path.exists(path):
        os.makedirs(path)

# Check whether this is within a py2exe zip
def get_exe_dir():
    import imp, sys
    if (hasattr(sys, "frozen") or # new py2exe
            hasattr(sys, "importers") # old py2exe
            or imp.is_frozen("__main__")): # tools/freeze
        return os.path.dirname(sys.executable)
    return None

_datapath = get_exe_dir()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.model.upgrades import allUpgrades
from trosnoth.trosnothgui.common import button
from trosnoth.gui.framework import prompt
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.framework.elements import TextElement
from trosnoth.gui.common import ScaledLocation, ScaledArea
from trosnoth.gui import keyboard
from trosnoth import keymap
from trosnoth.data import getPath, user
from trosnoth.utils.event import Event

class KeymapTab(Tab):

    def __init__(self, app, onClose=None):
        super(KeymapTab, self).__init__(app, 'Controls')
        self.font = app.screenManager.fonts.bigMenuFont

        self.onClose = Event()
        if onClose is not None:
            self.onClose.addListener(onClose)

        # Break things up into categories
        movement = ['jump', 'down', 'left', 'right']
        menus =    ['menu', 'more actions']
        actions =  ['respawn', 'select upgrade', 'activate upgrade',
                    'abandon', 'change nickname', 'ready']
        misc =     ['chat', 'follow']
        upgrades = [upgradeClass.action for upgradeClass in sorted(allUpgrades,
                key=lambda upgradeClass: upgradeClass.order)]
        upgrades.append('no upgrade')

        display =  ['leaderboard', 'toggle interface', 'toggle terminal']

        actionNames = {
            'abandon': 'Abandon upgrade',
            'select upgrade': 'Select upgrade',
            'activate upgrade' : 'Activate upgrade',
            'change nickname': 'Change nickname',
            'chat': 'Chat',
            'down': 'Drop down',
            'follow': 'Auto pan (replay)',
            'jump': 'Jump',
            'leaderboard': 'Show leaderboard',
            'left': 'Move left',
            'menu': 'Main menu',
            'more actions': 'Advanced',
            'no upgrade': 'Deselect upgrade',
            'ready': 'Toggle ready',
            'respawn': 'Respawn',
            'right': 'Move right',
            'status bar': 'Status bar',
            'timer': 'Show timer',
            'toggle interface': 'Toggle HUD',
            'toggle terminal': 'Toggle terminal',
            'zone progress': 'Show zone bar',
        }
        actionNames.update((upgradeClass.action, upgradeClass.name) for
                upgradeClass in allUpgrades)

        # Organise the categories by column
        self.layout = [
            [movement, menus],
            [actions, display],
            [upgrades, misc],
        ]

        self.errorInfo = TextElement(self.app, '', self.font,
                                 ScaledLocation(512, 580, 'center'))
        self.text = [self.errorInfo]
        self.inputLookup = {}
        xPos = 190

        # Lay everything out automatically.
        keymapFont = self.app.screenManager.fonts.keymapFont
        keymapInputFont = self.app.screenManager.fonts.keymapInputFont
        for column in self.layout:               # Each column
            yPos = 200
            for category in column:         # Each category
                for action in category:     # Each action
                    # Draw action name (eg. Respawn)
                    self.text.append(TextElement(self.app, actionNames[action],
                                    keymapFont,
                                    ScaledLocation(xPos, yPos+6,'topright'),
                                    self.app.theme.colours.headingColour
                                    ))

                    # Create input box
                    box = prompt.KeycodeBox(self.app, ScaledArea(xPos+10, yPos,
                            100, 30),
                                    font = keymapInputFont)
                    box.onClick.addListener(self.setFocus)
                    box.onChange.addListener(self.inputChanged)
                    box.__action = action
                    self.inputLookup[action] = box

                    yPos += 35  # Between items
                yPos += 35      # Between categories
            xPos += 320         # Between columns

        self.elements = self.text + self.inputLookup.values() + [
            button(app, 'restore default controls', self.restoreDefaults, (0,
                -125), 'midbottom', secondColour=app.theme.colours.white),
            button(app, 'save', self.saveSettings, (-100, -75), 'midbottom',
                secondColour=app.theme.colours.white),
            button(app, 'cancel', self.cancel, (100, -75), 'midbottom',
                secondColour=app.theme.colours.white),
        ]

        self.populateInputs()

    def inputChanged(self, box):
        # Remove the old key.
        try:
            oldKey = self.keyMapping.getkey(box.__action)
        except KeyError:
            pass
        else:
            del self.keyMapping[oldKey]

        # Set the new key.
        self.keyMapping[box.value] = box.__action

        # Refresh the display.
        self.refreshInputs()

    def populateInputs(self):
        # Set up the keyboard mapping.
        self.keyMapping = keyboard.KeyboardMapping(keymap.default_game_keys)

        try:
            # Try to load keyboard mappings from the user's personal settings.
            config = open(getPath(user, 'keymap'), 'rU').read()
            self.keyMapping.load(config)
        except IOError:
            pass

        # Refresh the display.
        self.refreshInputs()

    def refreshInputs(self):
        for column in self.layout:
            for category in column:
                for action in category:
                    # Get the current key and put it in the box.
                    try:
                        key = self.keyMapping.getkey(action)
                    except KeyError:
                        key = None
                    self.inputLookup[action].value = key

                    # Make the box white
                    self.inputLookup[action].backColour = (255, 255, 255)

    def restoreDefaults(self):
        self.keyMapping = keyboard.KeyboardMapping(keymap.default_game_keys)
        self.refreshInputs()

        self.incorrectInput("Default controls restored: press 'save' to "
                "confirm", (0, 128, 0))

    def clearBackgrounds(self):
        for action in self.inputLookup:
            self.inputLookup[action].backColour = (255, 255, 255)
        self.setFocus(None)

    def saveSettings(self):
        # Perform the save.
        open(getPath(user, 'keymap'), 'w').write(self.keyMapping.save())

        emptyBoxes = []

        for box in self.inputLookup.itervalues():
            if box.value is None:
                emptyBoxes.append(box)

        if len(emptyBoxes) > 0:
            self.populateInputs()
            for box in emptyBoxes:
                box.backColour = self.app.theme.colours.invalidDataColour

            self.incorrectInput('Warning: some actions have no key',
                    (192, 0, 0))
        else:
            self.mainMenu()

    def incorrectInput(self, string, colour):
        self.errorInfo.setColour(colour)
        self.errorInfo.setText(string)
        self.errorInfo.setFont(self.font)

    def cancel(self):
        self.populateInputs()
        self.mainMenu()

    def mainMenu(self):
        self.incorrectInput('', (0, 0, 0))
        self.clearBackgrounds()
        self.onClose.execute()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework import framework, elements
from trosnoth.utils.twist import WeakCallLater

class StarGroup(framework.CompoundElement):
    def __init__(self, app):
        super(StarGroup, self).__init__(app)

    def _remove(self, starAn):
        self.elements.remove(starAn)
    def star(self, pos):
        self.elements.append(StarAnimation(self.app, pos, self))

class StarAnimation(elements.PictureElement):
    def __init__(self, app, pos, starGroup):
        pic = app.theme.sprites.star

        super(StarAnimation, self).__init__(app, pic, pos)

        self.pos = pos
        self.count = 0
        self.starGroup = starGroup
        WeakCallLater(0.1, self, 'advance')

    def advance(self):
        self.pos = (self.pos[0], self.pos[1] - 10)
        self.setPos(self.pos)
        self.count += 1
        if self.count < 5:
            WeakCallLater(0.1, self, 'advance')
        else:
            self.starGroup._remove(self)



import random
import logging
import string

from trosnoth.utils import math

log = logging.getLogger('zone')


class ZoneDef(object):
    '''Stores static information about the zone.

    Attributes:
        adjacentZones - mapping from adjacent ZoneDef objects to collection of
            map blocks joining the zones.
        id - the zone id
        initialOwner - the team that initially owned this zone,
            represented as a number (0 for neutral, 1 or 2 for a team).
        pos - the coordinates of this zone in the map
    '''

    def __init__(self, id, initialOwnerIndex, xCoord, yCoord):
        self.adjacentZones = {}
        self.id = id
        self.initialOwnerIndex = initialOwnerIndex
        self.pos = xCoord, yCoord
        self.bodyBlocks = []
        self.interfaceBlocks = []

        primes, index = divmod(id, 26)
        self.label = string.uppercase[index] + "'" * primes

    ##
    # Return a tuple of adjacent zones, and whether each connection is open
    def zones_AdjInfo(self):
        ret = []
        for adj, blocks in self.adjacentZones.iteritems():
            ret.append((adj, not blocks[0].blocked))
        return tuple(ret)

    def __str__(self):
        if self.id is None:
            return 'Z---'
        return 'Z%3d' % self.id

    def randomPosition(self):
        '''
        Returns a random position in this zone.
        '''
        if random.random() * 3 > 1:
            return random.choice(self.bodyBlocks).randomPosition()
        return random.choice(self.interfaceBlocks).randomPositionInZone(self)


class ZoneState(object):
    '''Represents information about a given zone and its state.

    Attributes:
        universe
        dark - boolean
        owner - a team
        players - a Group
        turretedPlayer - None or a Player

    Attributes which should (eventually) be completely migrated to the zoneDef:
        id
    '''

    def __init__(self, universe, zoneDef):
        self.defn = zoneDef
        self.id = zoneDef.id

        universe.zoneWithDef[zoneDef] = self

        self.universe = universe
        teamIndex = zoneDef.initialOwnerIndex
        if teamIndex is None:
            self.owner = None
            self.dark = False
        else:
            self.owner = universe.teams[teamIndex]
            self.dark = True

            # Tell the team object that it owns one more zone
            self.owner.zoneGained()

        self.previousOwner = self.owner

        self.players = set()
        self.turretedPlayer = None
        self.frozen = False

    def __str__(self):
        # Debug: uniquely identify this zone within the universe.
        if self.id is None:
            return 'Z---'
        return 'Z%3d' % self.id

    def isNeutral(self):
        return self.owner is None

    def tag(self, player):
        '''This method should be called when the orb in this zone is tagged'''
        self.previousOwner = self.owner
        self.dark = False

        # Inform the team objects
        if self.owner:
            self.owner.zoneLost()
        if player is not None:
            team = player.team
            if team is not None:
                team.zoneGained()
            player.incrementStars()
        else:
            team = None

        self.owner = team
        for zone in self.getAdjacentZones():
            if zone.owner == team or self.isNeutral():
                # Allow the adjacent zone to check if it is entirely
                # surrounded by non-enemy zones
                zone.makeDarkIfNeeded()
        self.makeDarkIfNeeded()

    def updateByTrosballPosition(self, position):
        self.isDark = False
        if abs(position[0] - self.defn.pos[0]) < 1e-5:
            self.owner = None
        elif position[0] < self.defn.pos[0]:
            self.owner = self.universe.teams[1]
        else:
            self.owner = self.universe.teams[0]

    def isEnemyTeam(self, team):
        return team != self.owner and team is not None

    def setOwnership(self, team, dark):
        if self.owner is not None:
            self.owner.zoneLost()
        self.owner = team
        if team is not None:
            team.zoneGained()
        self.dark = dark

    def isDark(self):
        return self.dark

    def makeDarkIfNeeded(self):
        '''This method should be called by an adjacent friendly zone that has
        been tagged and gained. It will check to see if it now has all
        surrounding orbs owned by the same team, in which case it will
        change its zone ownership.'''

        # If the zone is already owned, ignore it
        if not self.dark and self.owner is not None:
            for zone in self.getAdjacentZones():
                if self.isEnemyTeam(zone.owner):
                    break
            else:
                self.dark = True

    def playerIsWithinTaggingDistance(self, player):
        tagDistance = 35
        distance = math.distance(self.defn.pos, player.pos)
        return distance < tagDistance

    def getContiguousZones(self):
        '''
        Returns a set of all contiguous zones with orb owned by the same team
        as this zone.
        '''
        owner = self.owner
        sector = set()
        stack = [self]
        while stack:
            zone = stack.pop()
            if zone in sector:
                continue
            sector.add(zone)
            for adjacentZone in zone.getAdjacentZones():
                if adjacentZone.owner == owner:
                    stack.append(adjacentZone)
        return sector

    def playerWhoTaggedThisZone(self):
        '''
        Checks to see whether the zone has been tagged. If not, return None.
        If it has, return a tuple of the player and team who tagged it.
        If the zone has been rendered Neutral, this will be None, None
        '''
        teamsToPlayers = self.getCountedPlayersByTeam()
        teamsWhoCanTag = self.teamsAbleToTag()
        taggingPlayers = []

        if (len(teamsWhoCanTag) == 1 and list(teamsWhoCanTag)[0] ==
                self.owner):
            # No need to check again - we are already the owner
            return

        for team in teamsWhoCanTag:
            for player in teamsToPlayers[team]:
                if self.playerIsWithinTaggingDistance(player):
                    taggingPlayers.append(player)
                    # Only allow one player from each team to have tagged it.
                    break
        if len(taggingPlayers) == 2:
            # Both teams tagged - becomes neutral
            return (None, None)
        elif (len(taggingPlayers) == 1 and list(taggingPlayers)[0].team !=
                self.owner):
            return (taggingPlayers[0], taggingPlayers[0].team)
        else:
            return None

    def clearPlayers(self):
        self.players.clear()

    def addPlayer(self, player):
        '''Adds a player to this zone'''
        self.players.add(player)

    def getCountedPlayersByTeam(self):
        result = dict((team, []) for team in self.universe.teams)

        for player in self.players:
            if player.dead or player.turret:
                # Turreted players do not count as a player for the purpose
                # of reckoning whether an enemy can capture the orb
                continue
            if player.team is not None:
                result[player.team].append(player)
        return result

    def getPlayerCounts(self):
        '''
        Returns a list of (count, teams) ordered by count descending, where
        count is the number of counted (living non-turret) players in the zone
        and teams is the teams with that number of players. Excludes teams
        which do not own the zone and cannot capture it.
        '''
        teamsByCount = {}
        for team, players in self.getCountedPlayersByTeam().iteritems():
            if (
                    team != self.owner
                    and not self.adjacentToAnotherZoneOwnedBy(team)):
                # If the enemy team doesn't have an adjacent zone, they don't
                # count towards numerical advantage.
                continue
            teamsByCount.setdefault(len(players), []).append(team)

        return sorted(teamsByCount.iteritems(), reverse=True)

    def isCapturableBy(self, team):
        '''
        Returns True or False depending on whether this zone can be captured by
        the given team. This takes into account both the zone location and the
        number of players in the zone.
        '''
        return team != self.owner and team in self.teamsAbleToTag()

    def isDisputed(self):
        '''
        Returns a value indicating whether this is a disputed zone. A disputed
        zone is defined as a zone which cannot be tagged by any enemy team, but
        could be if there was one more enemy player in the zone.
        '''
        moreThanThreeDefenders = False

        playerCounts = self.getPlayerCounts()
        while playerCounts:
            count, teams = playerCounts.pop(0)
            if moreThanThreeDefenders and count < 3:
                return False

            if count == 0:
                if any(t != self.owner for t in teams):
                    # There is a team which could tag if it had one attacker
                    return True

            elif count < 3:
                if len(teams) == 1:
                    # If it's an attacking team it's capturable, if it's a
                    # defending team it's not disputed.
                    return False

                # If an attacking team had one more player they could capture
                return True

            elif count == 3:
                if moreThanThreeDefenders:
                    return True

                if len(teams) == 1:
                    # If it's an attacking team it's capturable, if it's a
                    # defending team it's not disputed.
                    return False

                # Team could capture if it had 4 attackers
                return True

            else:
                if any(t != self.owner for t in teams):
                    # Already capturable
                    return False

                moreThanThreeDefenders = True

        return False

    def getAdjacentZones(self):
        '''
        Iterates through ZoneStates adjacent to this one.
        '''
        for adjZoneDef in self.defn.adjacentZones:
            yield self.universe.zoneWithDef[adjZoneDef]

    def getUnblockedNeighbours(self):
        '''
        Iterates through ZoneStates adjacent to this one which are not blocked
        off.
        '''
        for adjZoneDef, blocks in self.defn.adjacentZones.iteritems():
            if not blocks[0].blocked:
                yield self.universe.zoneWithDef[adjZoneDef]

    def adjacentToAnotherZoneOwnedBy(self, team):
        '''
        Returns whether or not this zone is adjacent to a zone whose orb is
        owned by the given team.
        '''
        for adjZone in self.getAdjacentZones():
            if adjZone.owner == team:
                return True
        return False

    def teamsAbleToTag(self):
        '''
        Returns the set of teams who have enough players to tag a zone
        (ignoring current zone ownership). Teams must have:
         (a) strict numerical advantage in this zone; or
         (b) more than 3 players in this zone.
        This takes into account the fact that turrets do not count towards
        numerical advantage.
        '''
        result = set()

        playerCounts = self.getPlayerCounts()
        while playerCounts:
            count, teams = playerCounts.pop(0)
            if count <= 3:
                if not result and len(teams) == 1 and count > 0:
                    result.update(teams)
                break
            result.update(teams)
        return result

    @staticmethod
    def canTag(numTaggers, numDefenders):
        '''
        Deprecated, do not use.
        '''
        return numTaggers > numDefenders or numTaggers > 3

    def consequenceOfCapture(self):
        '''
        Uses the zone neutralisation logic to calculate how many zone points an
        enemy team would gain by capturing this zone. That is, 2 points for the
        zone itself, plus one for each zone neutralised in the process.
        '''
        if self.owner is None:
            # Always one point for capturing a neutral zone
            return 1

        seen = {self}
        explore = [z for z in self.getAdjacentZones() if z.owner == self.owner]
        sectors = []
        while explore:
            zone = explore.pop(0)
            if zone in seen:
                continue

            thisSector = [zone]
            score = 0
            while thisSector:
                score += 1
                zone = thisSector.pop(0)
                seen.add(zone)
                for z in zone.getAdjacentZones():
                    if z.owner == self.owner and z not in seen:
                        thisSector.append(z)
            sectors.append(score)

        if sectors:
            # Largest sector is not lost
            sectors.remove(max(sectors))

        # Two points for capture, plus one for each zone neutralised
        return 2 + sum(sectors)



# coding: utf-8

from django.contrib.auth.models import User
from django.db import models
from django.utils.encoding import python_2_unicode_compatible


class TrosnothServerSettings(models.Model):
    serverName = models.TextField(default='My First Trosnoth Server')
    welcomeText = models.TextField(default=(
        'Congratulations! You have successfully installed your Trosnoth '
        'server. <a href="admin/trosnoth/trosnothserversettings/">'
        'Click here</a> to configure it.'))

    allowRemoteGameRegistration = models.BooleanField(default=True)

    iceEnabled = models.BooleanField(default=False)
    iceHost = models.TextField(default='127.0.0.1')
    icePort = models.IntegerField(default=6502)
    iceProxyStringOverride = models.TextField(default='', blank=True)
    iceSecret = models.TextField(default='', blank=True)

    @staticmethod
    def get():
        rows = TrosnothServerSettings.objects.all()
        if rows.count() == 0:
            result = TrosnothServerSettings()
            result.save()
        else:
            result = rows[0]
        return result

    class Meta:
        verbose_name_plural = 'Trosnoth server setting'


class TrosnothUser(models.Model):
    user = models.OneToOneField(
        User, on_delete=models.CASCADE, primary_key=True)
    nick = models.TextField(unique=True)
    lastSeen = models.DateTimeField(null=True, blank=True)
    oldPasswordHash = models.BinaryField(default='')

    def __str__(self):
        return '{} ({})'.format(self.nick, self.user.username)

    @staticmethod
    def fromUser(**kwargs):
        user = User.objects.get(**kwargs)
        if hasattr(user, 'trosnothuser'):
            result = user.trosnothuser
        else:
            result = TrosnothUser(user=user, nick=user.username)
            result.save()
        return result


class AchievementProgress(models.Model):
    user = models.ForeignKey(TrosnothUser)
    achievementId = models.TextField()
    unlocked = models.BooleanField(default=False)
    progress = models.IntegerField(default=0)
    data = models.TextField(blank=True)

    class Meta:
        verbose_name_plural = 'Achievement progress records'
        unique_together = ('user', 'achievementId')

    def __str__(self):
        return '{}: {}'.format(self.user.nick, self.achievementId)


@python_2_unicode_compatible
class GameRecord(models.Model):
    started = models.DateTimeField()
    finished = models.DateTimeField()
    gameSeconds = models.FloatField(default=0)
    mapHeight = models.IntegerField()
    halfMapWidth = models.IntegerField()
    serverVersion = models.TextField()
    blueTeamName = models.TextField()
    redTeamName = models.TextField()
    winningTeam = models.CharField(max_length=1, blank=True)
    replayName = models.TextField(default='', blank=True)

    def __str__(self):
        secs = self.gameSeconds
        mins, secs = divmod(secs, 60)
        if mins == 0:
            duration = '{} s'.format(secs)
        else:
            duration = '{}:{:02d}'.format(int(mins), int(secs))

        return u'Game {} ({} vs. {}, {}, {})'.format(
            self.pk,
            self.blueTeamName, self.redTeamName,
            self.getScoreString(),
            duration,
        )

    def getScoreString(self):
        if not self.winningTeam:
            scores = u'½-½'
        elif self.winningTeam == 'A':
            scores = '1-0'
        else:
            scores = '0-1'
        return scores


class GamePlayer(models.Model):
    game = models.ForeignKey(GameRecord)
    user = models.ForeignKey(TrosnothUser, null=True)
    bot = models.BooleanField(default=False)
    botName = models.TextField(blank=True, default='')
    team = models.CharField(max_length=1, blank=True)

    starsEarned = models.IntegerField(default=0)
    starsWasted = models.IntegerField(default=0)
    starsUsed = models.IntegerField(default=0)
    kills = models.IntegerField(default=0)
    deaths = models.IntegerField(default=0)
    zoneTags = models.IntegerField(default=0)
    zoneAssists = models.IntegerField(default=0)
    shotsFired = models.IntegerField(default=0)
    shotsHit = models.IntegerField(default=0)
    timeAlive = models.FloatField(default=0)
    timeDead = models.FloatField(default=0)
    killStreak = models.IntegerField(default=0)
    tagStreak = models.IntegerField(default=0)
    aliveStreak = models.FloatField(default=0)

    class Meta:
        unique_together = ('game', 'user')

    def __str__(self):
        return '{}: {}'.format(self.nameStr(), self.game)

    def nameStr(self):
        if self.user:
            return '{}'.format(self.user)
        if self.bot:
            return '{} (bot)'.format(self.botName)
        return '{} (unregistered)'.format(self.botName)


class UpgradesUsedInGameRecord(models.Model):
    gamePlayer = models.ForeignKey(GamePlayer)
    upgrade = models.CharField(max_length=1)
    count = models.IntegerField(default=0)

    class Meta:
        unique_together = ('gamePlayer', 'upgrade')

    def __str__(self):
        return '{}: {}: {}'.format(self.gamePlayer, self.upgrade, self.count)


class PlayerKills(models.Model):
    killer = models.ForeignKey(GamePlayer, related_name='+', null=True)
    killee = models.ForeignKey(GamePlayer, related_name='+')
    count = models.IntegerField(default=0)

    class Meta:
        verbose_name_plural = 'Player kills records'
        unique_together = ('killer', 'killee')

    def __str__(self):
        return '{}: {} killed {}: {}'.format(
            self.getGame(),
            self.killer.nameStr() if self.killer else 'no-one',
            self.killee.nameStr(),
            self.count,
        )

    def getGame(self):
        return self.killee.game



from functools import partial
import logging
import random
import pygame

from trosnoth.const import (
    TROSBALL_DEATH, OFF_MAP_DEATH, OPEN_CHAT, PRIVATE_CHAT, TEAM_CHAT,
    NOT_ENOUGH_STARS_REASON, HAVE_UPGRADE_REASON, PLAYER_DEAD_REASON,
    GAME_NOT_STARTED_REASON, ALREADY_TURRET_REASON, TOO_CLOSE_TO_EDGE_REASON,
    TOO_CLOSE_TO_ORB_REASON, NOT_IN_DARK_ZONE_REASON, INVALID_UPGRADE_REASON,
    DISABLED_UPGRADE_REASON, ALREADY_ALIVE_REASON, BE_PATIENT_REASON,
    ENEMY_ZONE_REASON, FROZEN_ZONE_REASON, TICK_PERIOD,
)
from trosnoth.gui.framework import framework, hotkey, console
from trosnoth.gui.framework.elements import (
    TextElement, SolidRect, TextButton,
)
from trosnoth.gui.framework.collapsebox import CollapseBox
from trosnoth.gui import keyboard
from trosnoth.gui.common import (
    Region, Screen, Location, Canvas, PaddedRegion, ScaledScalar,
    ScaledLocation,
)

from trosnoth.gamerecording.achievementlist import availableAchievements

from trosnoth.model.agent import ConcreteAgent

from trosnoth.trosnothgui.ingame import viewManager
from trosnoth.trosnothgui.ingame.replayInterface import ViewControlInterface
from trosnoth.trosnothgui.ingame.joincontroller import JoinGameController
from trosnoth.trosnothgui.ingame.detailsInterface import DetailsInterface
from trosnoth.trosnothgui.ingame.playerInterface import PlayerInterface

from trosnoth import keymap

from trosnoth.data import user, getPath

from trosnoth.utils import globaldebug
from trosnoth.utils.math import distance
from trosnoth.utils.event import Event
from trosnoth.utils.twist import WeakCallLater

from trosnoth.messages import (
    TaggingZoneMsg, ChatFromServerMsg, ChatMsg, GameStartMsg, GameOverMsg,
    StartingSoonMsg, ShotFiredMsg, RespawnMsg, CannotRespawnMsg, TickMsg,
    CannotJoinMsg, AddPlayerMsg, PlayerHasUpgradeMsg, RemovePlayerMsg,
    PlayerStarsSpentMsg, CannotBuyUpgradeMsg, ConnectionLostMsg,
    AchievementUnlockedMsg, SetPlayerTeamMsg,
    PlayerHasElephantMsg, FireShoxwaveMsg,
    PlayerHasTrosballMsg, TrosballTagMsg, UpgradeChangedMsg,
)

from twisted.internet import defer

log = logging.getLogger(__name__)


class GameInterface(framework.CompoundElement, ConcreteAgent):
    '''Interface for when we are connected to a game.'''

    achievementDefs = availableAchievements

    def __init__(
            self, app, game, onDisconnectRequest=None,
            onConnectionLost=None, replay=False, authTag=0):
        super(GameInterface, self).__init__(app, game=game)
        self.localState.onShoxwave.addListener(self.localShoxwaveFired)
        self.world.onSectorNeutralised.addListener(self.sectorNeutralised)
        self.world.onOpenChatReceived.addListener(self.openChat)
        self.world.onTeamChatReceived.addListener(self.teamChat)
        self.world.onReset.addListener(self.worldReset)
        self.world.onGrenadeExplosion.addListener(self.grenadeExploded)
        self.world.onTrosballExplosion.addListener(self.trosballExploded)
        self.world.onBomberExplosion.addListener(self.trosballExploded)
        self.world.onGameStateChanged.addListener(self.gameStateChanged)

        self.authTag = authTag

        self.subscribedPlayers = set()

        self.onDisconnectRequest = Event()
        if onDisconnectRequest is not None:
            self.onDisconnectRequest.addListener(onDisconnectRequest)

        self.onConnectionLost = Event()
        if onConnectionLost is not None:
            self.onConnectionLost.addListener(onConnectionLost)
        self.game = game

        self.joinDialogReason = None

        self.keyMapping = keyboard.KeyboardMapping(keymap.default_game_keys)
        self.updateKeyMapping()
        self.gameViewer = viewManager.GameViewer(self.app, self, game, replay)
        if replay:
            self.joinController = None
        else:
            self.joinController = JoinGameController(self.app, self, self.game)
        self.detailsInterface = DetailsInterface(self.app, self)
        self.runningPlayerInterface = None
        self.winnerMsg = WinnerMsg(app)
        self.timingInfo = TimingInfo(
            app, self, Location(Canvas(307, 768), 'midbottom'),
            app.screenManager.fonts.consoleFont,
            )
        self.gameInfoDisplay = GameInfoDisplay(
            app, self,
            Region(topleft=Screen(0.01, 0.05), size=Canvas(330, 200)))
        self.hotkeys = hotkey.Hotkeys(
            self.app, self.keyMapping, self.detailsInterface.doAction)
        self.terminal = None

        self.joinInProgress = False

        self.vcInterface = None
        if replay:
            self.vcInterface = ViewControlInterface(self.app, self)

        self.ready = False
        defer.maybeDeferred(game.addAgent, self).addCallback(self.addedAgent)

        self.setElements()

    def gameStateChanged(self):
        self.gameInfoDisplay.refreshInfo()
        self.winnerMsg.hide()

    def addedAgent(self, result):
        self.ready = True
        if self.joinController:
            self.joinDialogReason = 'automatic'
            self.joinController.start()

    def spectatorWantsToJoin(self):
        if self.runningPlayerInterface or not self.joinController:
            return
        self.joinDialogReason = 'from menu'
        self.joinController.maybeShowJoinDialog(autoJoin=True)

    def sendRequest(self, msg):
        if not self.ready:
            # Not yet completely connected to game
            return
        super(GameInterface, self).sendRequest(msg)

    def worldReset(self, *args, **kwarsg):
        self.winnerMsg.hide()
        if self.ready and self.joinController:
            self.joinController.gotWorldReset()
        self.gameViewer.reset()

    def updateKeyMapping(self):
        # Set up the keyboard mapping.
        try:
            # Try to load keyboard mappings from the user's personal settings.
            config = open(getPath(user, 'keymap'), 'rU').read()
            self.keyMapping.load(config)
        except IOError:
            pass

    @ConnectionLostMsg.handler
    def connectionLost(self, msg):
        self.cleanUp()
        if self.joinController:
            self.joinController.hide()
        self.onConnectionLost.execute()

    def joined(self, player):
        '''Called when joining of game is successful.'''
        pygame.key.set_repeat()
        self.gameViewer.worldgui.overridePlayer(self.localState.player)
        self.runningPlayerInterface = pi = PlayerInterface(self.app, self)
        self.detailsInterface.setPlayer(pi.player)
        self.setElements()

        self.joinController.hide()
        self.gameViewer.leaderboard.update()

    def spectate(self):
        '''
        Called by join controller if user selects to only spectate.
        '''
        self.vcInterface = ViewControlInterface(self.app, self)
        self.setElements()
        self.joinController.hide()

    def joinDialogCancelled(self):
        if self.joinDialogReason == 'automatic':
            self.disconnect()
        else:
            self.spectate()

    def stop(self):
        super(GameInterface, self).stop()
        self.localState.onShoxwave.removeListener(self.localShoxwaveFired)
        self.world.onSectorNeutralised.removeListener(self.sectorNeutralised)
        self.world.onOpenChatReceived.removeListener(self.openChat)
        self.world.onTeamChatReceived.removeListener(self.teamChat)
        self.world.onReset.removeListener(self.worldReset)
        self.world.onGrenadeExplosion.removeListener(self.grenadeExploded)
        self.world.onTrosballExplosion.removeListener(self.trosballExploded)
        self.world.onBomberExplosion.removeListener(self.trosballExploded)
        self.world.onGameStateChanged.removeListener(self.gameStateChanged)
        self.gameViewer.stop()
        if self.runningPlayerInterface is not None:
            self.runningPlayerInterface.stop()

    def setElements(self):
        spectate = replay = False
        if self.runningPlayerInterface:
            self.elements = [
                self.gameViewer, self.runningPlayerInterface,
                self.gameInfoDisplay, self.hotkeys, self.detailsInterface,
                self.winnerMsg, self.timingInfo]
        else:
            self.elements = [
                self.gameViewer, self.gameInfoDisplay,
                self.hotkeys, self.detailsInterface,
                self.winnerMsg, self.timingInfo]
            if self.vcInterface is not None:
                self.elements.insert(2, self.vcInterface)

            if self.joinController:
                spectate = True
            else:
                replay = True
        self.detailsInterface.menuManager.setMode(
            spectate=spectate, replay=replay)

    def toggleTerminal(self):
        if self.terminal is None:
            locs = {'app': self.app}
            if hasattr(self.app, 'getConsoleLocals'):
                locs.update(self.app.getConsoleLocals())
            self.terminal = console.TrosnothInteractiveConsole(
                self.app,
                self.app.screenManager.fonts.consoleFont,
                Region(size=Screen(1, 0.4), bottomright=Screen(1, 1)),
                locals=locs)
            self.terminal.interact().addCallback(self._terminalQuit)

        from trosnoth.utils.utils import timeNow
        if self.terminal in self.elements:
            if timeNow() > self._termWaitTime:
                self.elements.remove(self.terminal)
        else:
            self._termWaitTime = timeNow() + 0.1
            self.elements.append(self.terminal)
            self.setFocus(self.terminal)

    def _terminalQuit(self, result):
        if self.terminal in self.elements:
            self.elements.remove(self.terminal)
        self.terminal = None

    def disconnect(self):
        self.cleanUp()
        self.onDisconnectRequest.execute()

    def joinGame(self, nick, team, timeout=10):
        if self.joinInProgress:
            return

        if team is None:
            teamId = '\x00'
        else:
            teamId = team.id

        self.joinInProgress = True
        self.sendJoinRequest(teamId, nick, authTag=self.authTag)
        WeakCallLater(timeout, self, '_joinTimedOut')

    def setPlayer(self, player):
        if not player:
            self.gameViewer.worldgui.removeOverride()
            self.lostPlayer()

        super(GameInterface, self).setPlayer(player)

        if player:
            if __debug__ and globaldebug.enabled:
                globaldebug.localPlayerId = player.id

            self.joinInProgress = False
            self.joined(player)

    @CannotJoinMsg.handler
    def joinFailed(self, msg):
        self.joinInProgress = False
        self.joinController.joinFailed(msg.reasonId)

    def _joinTimedOut(self):
        if self.player or not self.joinInProgress:
            return
        self.joinInProgress = False
        self.joinController.joinFailed('timeout')

    def cleanUp(self):
        if self.gameViewer.timerBar is not None:
            self.gameViewer.timerBar.kill()
            self.gameViewer.timerBar = None
        pygame.key.set_repeat(300, 30)

    def gameOver(self, winningTeam):
        if winningTeam:
            self.winnerMsg.show(
                '%s win' % (winningTeam,),
                self.app.theme.colours.chatColour(winningTeam),
            )

            self.playSound('gameLose')
        else:
            self.winnerMsg.show('Game drawn', (128, 128, 128))

    @PlayerStarsSpentMsg.handler
    def discard(self, msg):
        pass

    @PlayerHasElephantMsg.handler
    def gotElephant(self, msg, _lastElephantPlayer=[None]):
        player = self.world.getPlayer(msg.playerId)
        if player and player != _lastElephantPlayer[0]:
            message = '%s now has Jerakeen!' % (player.nick,)
            self.detailsInterface.newMessage(message)
            _lastElephantPlayer[0] = player

    @PlayerHasTrosballMsg.handler
    def gotTrosball(self, msg, _lastTrosballPlayer=[None]):
        player = self.world.playerWithId.get(msg.playerId)

        if player != _lastTrosballPlayer[0]:
            _lastTrosballPlayer[0] = player
            if player is None:
                message = 'The ball has been dropped!'
            else:
                message = '%s has the ball!' % (player.nick,)
            self.detailsInterface.newMessage(message)

    @AddPlayerMsg.handler
    def addPlayer(self, msg):
        player = self.world.getPlayer(msg.playerId)
        if player and player not in self.subscribedPlayers:
            self.subscribedPlayers.add(player)
            team = player.team if player.team else self.world.rogueTeamName
            message = '%s has joined %s' % (player.nick, team)
            self.detailsInterface.newMessage(message)
            player.onDied.addListener(partial(self.playerDied, player))

    @SetPlayerTeamMsg.handler
    def changeTeam(self, msg):
        self.defaultHandler(msg)    # Make sure the local player changes team
        player = self.world.getPlayer(msg.playerId)
        if player:
            message = '%s has joined %s' % (
                player.nick, self.world.getTeamName(msg.teamId))
            self.detailsInterface.newMessage(message)

    @RemovePlayerMsg.handler
    def handle_RemovePlayerMsg(self, msg):
        player = self.world.getPlayer(msg.playerId)
        if player:
            message = '%s has left the game' % (player.nick,)
            self.detailsInterface.newMessage(message)
            self.subscribedPlayers.discard(player)

    def lostPlayer(self):
        self.runningPlayerInterface.stop()
        self.runningPlayerInterface = None
        self.setElements()

    @GameStartMsg.handler
    def gameStart(self, msg):
        message = 'The game is now on!!'
        self.detailsInterface.newMessage(message)
        if self.gameViewer.timerBar:
            self.gameViewer.timerBar.loop()
        self.playSound('startGame', 1)

    @GameOverMsg.handler
    def handle_GameOverMsg(self, msg):
        super(GameInterface, self).handle_GameOverMsg(msg)
        try:
            team = self.world.teamWithId[msg.teamId]
        except KeyError:
            team = None
        self.gameOver(team)
        if msg.timeOver:
            message2 = 'Game time limit has expired'
            self.detailsInterface.newMessage(message2)

    @TrosballTagMsg.handler
    def trosballScore(self, msg):
        try:
            team = self.world.teamWithId[msg.teamId]
        except KeyError:
            return

        if team is None:
            # Resetting Trosball
            return

        player = self.world.playerWithId.get(msg.playerId)
        if player is not None:
            if player.team == team:
                message = '%s scores for %s!' % (player.nick, team.teamName)
                bigMessage = '%s scored!' % player.nick
                self.playSound('goal', 1)
            else:
                message = '%s scores an own goal!' % (player.nick,)
                if self.detailsInterface.player.id == player.id:
                    bigMessage = '(Psst... wrong goal!)'
                else:
                    bigMessage = 'Goal!'
                self.playSound('ownGoal', 1)
        else:
            message = '%s scores!' % (team.teamName,)
            bigMessage = 'Goal!'
            self.playSound('goal', 1)

        self.winnerMsg.show(
            bigMessage, self.app.theme.colours.chatColour(team))

        self.detailsInterface.newMessage(message)
        teams = list(self.world.teams)
        message = '%s: %d - %s: %d' % (
            teams[0].teamName,
            self.world.trosballScore[teams[0].id],
            teams[1].teamName,
            self.world.trosballScore[teams[1].id],
        )
        self.detailsInterface.newMessage(message)

    def getChatColour(self, team):
        return self.app.theme.colours.chatColour(team)

    @CannotBuyUpgradeMsg.handler
    def notEnoughStars(self, msg):
        if msg.reasonId == NOT_ENOUGH_STARS_REASON:
            text = 'Your team does not have enough stars.'
        elif msg.reasonId == HAVE_UPGRADE_REASON:
            text = 'You already have an upgrade.'
        elif msg.reasonId == PLAYER_DEAD_REASON:
            text = 'You cannot buy an upgrade while dead.'
        elif msg.reasonId == GAME_NOT_STARTED_REASON:
            text = 'Upgrades can''t be bought at this time.'
        elif msg.reasonId == ALREADY_TURRET_REASON:
            text = 'There is already a turret in this zone.'
        elif msg.reasonId == TOO_CLOSE_TO_EDGE_REASON:
            text = 'You are too close to the zone edge.'
        elif msg.reasonId == TOO_CLOSE_TO_ORB_REASON:
            text = 'You are too close to the orb.'
        elif msg.reasonId == NOT_IN_DARK_ZONE_REASON:
            text = 'You are not in a dark friendly zone.'
        elif msg.reasonId == INVALID_UPGRADE_REASON:
            text = 'Upgrade not recognised by server.'
        elif msg.reasonId == DISABLED_UPGRADE_REASON:
            text = 'That upgrade is currently disabled.'
        else:
            text = 'You cannot buy that upgrade at this time.'
        self.detailsInterface.newMessage(text)
        self.defaultHandler(msg)

    @PlayerHasUpgradeMsg.handler
    def gotUpgrade(self, msg):
        player = self.world.getPlayer(msg.playerId)
        if player:
            self.detailsInterface.upgradeUsed(player)

            if (self.detailsInterface.player is None or
                    self.detailsInterface.player.isFriendsWith(player)):
                if msg.upgradeType == 't':
                    self.playSound('turret')
                else:
                    self.playSound('buyUpgrade')
        self.defaultHandler(msg)

    @ChatFromServerMsg.handler
    def gotChatFromServer(self, msg):
        self.detailsInterface.newMessage(msg.text.decode(), error=msg.error)

    @StartingSoonMsg.handler
    def startingSoon(self, msg):
        self.detailsInterface.newMessage(
            'Game starting in %d seconds' % (msg.delay,))

    @TaggingZoneMsg.handler
    def zoneTagged(self, msg):
        try:
            zone = self.world.zoneWithId[msg.zoneId]
            zoneLabel = zone.defn.label
        except KeyError:
            zoneLabel = '<?>'

        if msg.playerId != '\x00':
            try:
                player = self.world.playerWithId[msg.playerId]
            except KeyError:
                nick = '<?>'
            else:
                nick = player.nick
                # Draw a star.
                rect = self.gameViewer.worldgui.getPlayerSprite(player.id).rect
                self.detailsInterface.starGroup.star(rect.center)
            message = '%s tagged zone %s' % (nick, zoneLabel)

            self.detailsInterface.newMessage(message)

    def sectorNeutralised(self, tagger, zoneCount):
        theWordZone = 'zone' if zoneCount == 1 else 'zones'
        if not tagger:
            message = '%d %s neutralised!' % (zoneCount, theWordZone)
        else:
            nick = tagger.nick
            message = '%s neutralised %d %s!' % (
                nick, zoneCount, theWordZone)
        self.detailsInterface.newMessage(message)

    def playerDied(self, target, killer, deathType):
        if deathType == OFF_MAP_DEATH:
            messages = [
                'fell into the void', 'looked into the abyss',
                'dug too greedily and too deep']
            message = '%s %s' % (target.nick, random.choice(messages))
        elif deathType == TROSBALL_DEATH:
            message = '%s was killed by the Trosball' % (target.nick,)
        else:
            if killer is None:
                message = '%s was killed' % (target.nick,)
                self.detailsInterface.newMessage(message)
            else:
                message = '%s killed %s' % (killer.nick, target.nick)

            if (self.runningPlayerInterface is not None and
                    self.runningPlayerInterface.player == killer):
                rect = self.gameViewer.worldgui.getPlayerSprite(
                    target.id).rect
                self.detailsInterface.starGroup.star(rect.center)
        self.detailsInterface.newMessage(message)

    @RespawnMsg.handler
    def playerRespawn(self, msg):
        player = self.world.getPlayer(msg.playerId)
        if player:
            message = '%s is back in the game' % (player.nick,)
            self.detailsInterface.newMessage(message)

    @CannotRespawnMsg.handler
    def respawnFailed(self, msg):
        if msg.reasonId == GAME_NOT_STARTED_REASON:
            message = 'The game has not started yet.'
        elif msg.reasonId == ALREADY_ALIVE_REASON:
            message = 'You are already alive.'
        elif msg.reasonId == BE_PATIENT_REASON:
            message = 'You cannot respawn yet.'
        elif msg.reasonId == ENEMY_ZONE_REASON:
            message = 'Cannot respawn outside friendly zone.'
        elif msg.reasonId == FROZEN_ZONE_REASON:
            message = 'That zone has been frozen!'
        else:
            message = 'You cannot respawn here.'
        self.detailsInterface.newMessage(
            message, self.app.theme.colours.errorMessageColour)

    def sendPrivateChat(self, player, targetId, text):
        self.sendRequest(ChatMsg(PRIVATE_CHAT, targetId, text=text.encode()))

    def sendTeamChat(self, player, text):
        self.sendRequest(
            ChatMsg(TEAM_CHAT, player.teamId, text=text.encode()))

    def sendPublicChat(self, player, text):
        self.sendRequest(ChatMsg(OPEN_CHAT, text=text.encode()))

    def openChat(self, text, sender):
        text = ': ' + text
        self.detailsInterface.newChat(text, sender)

    def teamChat(self, team, text, sender):
        player = self.detailsInterface.player
        if player and player.isFriendsWithTeam(team):
            text = " (team): " + text
            self.detailsInterface.newChat(text, sender)

    @AchievementUnlockedMsg.handler
    def achievementUnlocked(self, msg):
        player = self.world.getPlayer(msg.playerId)
        if not player:
            return

        achievementName = self.achievementDefs.getAchievementDetails(
            msg.achievementId)[0]
        self.detailsInterface.newMessage(
            '%s has unlocked "%s"!' % (player.nick, achievementName),
            self.app.theme.colours.achievementMessageColour)

        focusPlayer = self.detailsInterface.player
        if (focusPlayer is not None and focusPlayer.id == msg.playerId):
            self.detailsInterface.localAchievement(msg.achievementId)

    @ShotFiredMsg.handler
    def shotFired(self, msg):
        self.defaultHandler(msg)
        try:
            shot = self.world.getShot(msg.shotId)
        except KeyError:
            return

        pos = shot.pos
        dist = self.distance(pos)
        self.playSound('shoot', self.getSoundVolume(dist))

    def grenadeExploded(self, pos, radius):
        self.gameViewer.worldgui.addExplosion(pos)
        dist = self.distance(pos)
        self.playSound('explodeGrenade', self.getSoundVolume(dist))

    def trosballExploded(self, player):
        self.gameViewer.worldgui.addTrosballExplosion(player.pos)
        dist = self.distance(player.pos)
        self.playSound('explodeGrenade', self.getSoundVolume(dist))

    @FireShoxwaveMsg.handler
    def shoxwaveExplosion(self, msg):
        localPlayer = self.localState.player
        if localPlayer and msg.playerId == localPlayer.id:
            return
        self.gameViewer.worldgui.addShoxwaveExplosion((msg.xpos, msg.ypos))

    def localShoxwaveFired(self):
        localPlayer = self.localState.player
        self.gameViewer.worldgui.addShoxwaveExplosion(localPlayer.pos)

    @UpgradeChangedMsg.handler
    def upgradeChanged(self, msg):
        self.detailsInterface.upgradeDisplay.refresh()

    def distance(self, pos):
        return distance(self.gameViewer.viewManager.getTargetPoint(), pos)

    def getSoundVolume(self, distance):
        'The volume for something that far away from the player'
        # Up to 500px away is within the "full sound zone" - full sound
        distFromScreen = max(0, distance - 500)
        # 1000px away from "full sound zone" is 0 volume:
        return 1 - min(1, (distFromScreen / 1000.))

    def playSound(self, action, volume=1):
        self.app.soundPlayer.play(action, volume)

    @TickMsg.handler
    def handle_TickMsg(self, msg):
        super(GameInterface, self).handle_TickMsg(msg)
        self.timingInfo.ticksSeen += 1


class TimingInfo(framework.Element):
    def __init__(self, app, gameInterface, location, font, *args, **kwargs):
        super(TimingInfo, self).__init__(app, *args, **kwargs)
        self.interface = gameInterface
        self.location = location
        self.font = font
        self.framesSeen = 0
        self.ticksSeen = 0
        self.timePassed = 0.
        self.lastDelay = None
        self.make_image()

    def make_image(self):
        bits = []
        if self.timePassed > 0:
            bits.append(self.font.render(
                self.app,
                'FPS: %.1f' % (self.framesSeen / self.timePassed),
                True, (0, 0, 0),
            ))
            bits.append(self.font.render(
                self.app,
                'TPS: %.1f' % (self.ticksSeen / self.timePassed),
                True, (0, 0, 0),
            ))
        self.lastDelay = self.interface.localState.serverDelay
        delay = self.interface.localState.serverDelay * TICK_PERIOD
        bits.append(self.font.render(
            self.app, 'RTT: %s s' % (delay,), True, (0, 0, 0)))

        width = max(b.get_width() for b in bits) if bits else 0
        height = sum(b.get_height() for b in bits)
        self.image = pygame.Surface((width, height), pygame.SRCALPHA)

        y = 0
        for bit in bits:
            self.image.blit(bit, (0, y))
            y += bit.get_height()

        self.framesSeen = 0
        self.ticksSeen = 0
        self.timePassed = 0.

    def tick(self, deltaT):
        if not self.app.displaySettings.showTimings:
            return
        self.timePassed += deltaT

    def draw(self, screen):
        if not self.app.displaySettings.showTimings:
            return
        self.framesSeen += 1
        if (
                self.timePassed > 3
                or self.lastDelay != self.interface.localState.serverDelay):
            self.make_image()
        rect = self.image.get_rect()
        self.location.apply(self.app, rect)
        screen.blit(self.image, rect)


class GameInfoDisplay(CollapseBox):
    def __init__(self, app, gameInterface, region):
        colours = app.theme.colours
        fonts = app.screenManager.fonts
        self.interface = gameInterface
        super(GameInfoDisplay, self).__init__(
            app,
            region=region,
            titleFont=fonts.gameInfoTitleFont,
            font=fonts.gameInfoFont,
            titleColour=colours.gameInfoTitle,
            hvrColour=colours.gameInfoHover,
            colour=colours.gameInfoColour,
            backColour=colours.gameInfoBackColour,
            title='',
        )
        self.refreshCall = None
        self.refreshInfo()

    def refreshInfo(self):
        if self.refreshCall:
            self.refreshCall.cancel()
        world = self.interface.world
        player = self.interface.localState.player
        title, lines = world.state.getInfo(world, player)
        self.setInfo(lines, title)

        # Check every 3 seconds
        self.refreshCall = WeakCallLater(3, self, 'doRefreshCall')

    def doRefreshCall(self):
        self.refreshCall = None
        self.refreshInfo()


class WinnerMsg(framework.CompoundElement):
    def __init__(self, app):
        super(WinnerMsg, self).__init__(app)
        self.winnerMsg = TextElement(
            app, '', app.screenManager.fonts.winMessageFont,
            Location(Screen(0.5, 0.05), 'midtop'), (64, 64, 64))
        self.background = SolidRect(
            app, (128, 128, 128), 150,
            PaddedRegion(self.winnerMsg, ScaledScalar(15)))
        self.elements = []

    def show(self, text, colour):
        self.winnerMsg.setText(text)
        self.background.colour = colour
        self.background.border = colour
        self.background.refresh()
        self.elements = [self.background, self.winnerMsg]

    def hide(self):
        self.elements = []



import logging
import random

from twisted.internet import defer, reactor

from trosnoth.ai import makeAIAgent, listAIs
from trosnoth.const import (
    TICK_PERIOD, LAG_BUFFER, LAG_THRESHOLD, INITIAL_ASSUMED_LATENCY,
)
from trosnoth.model import gameStates
from trosnoth.model.hub import Node
from trosnoth.model.universe import Universe, ServerUniverse
from trosnoth.messages import (
    RemovePlayerMsg, ConnectionLostMsg, DelayUpdatedMsg,
    ResyncAcknowledgedMsg, PlayerNoticedZombieHitMsg,
)
from trosnoth.gamerecording.gamerecorder import GameRecorder
from trosnoth.gamerecording.achievements import AchievementManager
from trosnoth.utils import globaldebug
from trosnoth.utils.event import Event
from trosnoth.utils.twist import WeakLoopingCall

log = logging.getLogger('game')


class Game(object):
    '''
    The most important interface in the Trosnoth code base is the Game/Agent
    interface. A Game represents either a locally running server or a game that
    is running on a remote server. In both of these cases, the game will
    provide the same interface for agents to connect to for receiving orders
    from the game, sending requests, and examining the state of the Universe.
    '''

    def __init__(self, world, ais=False, *args, **kwargs):
        super(Game, self).__init__(*args, **kwargs)
        self.world = world

        self.onServerCommand = Event()      # (msg)

        if ais:
            self.aiInjector = AIInjector(self)
            reactor.callLater(0, self.aiInjector.start)
        else:
            self.aiInjector = None

        world.onPlayerRemoved.addListener(self.playerRemoved)

    def addAgent(self, agent):
        '''
        Connects an agent to this game. An agent may be a user interface,
        an AI player, or anything that wants to receive events from the game
        and potentially send actions to it.
        Every player in Trosnoth must be controlled by an agent and no agent
        can control more than one player.
        May return a deferred which will fire once the agent is successfully
        added.
        '''
        raise NotImplementedError('%s.addAgent' % (self.__class__.__name__,))

    @defer.inlineCallbacks
    def addAI(self, aiName, team=None):
        ai = makeAIAgent(self, aiName)
        yield self.addAgent(ai)

        if team is not None:
            team = self.world.teams[team]
        ai.start(team)
        defer.returnValue(ai)

    def detachAgent(self, agent):
        raise NotImplementedError('%s.detachAgent' % (
            self.__class__.__name__,))

    def agentRequest(self, agent, msg):
        '''
        Called by an agent with an AgentRequest message that it wants to
        request.
        '''
        raise NotImplementedError('%s.agentRequest' % (
            self.__class__.__name__,))

    def stop(self):
        if self.aiInjector is not None:
            self.aiInjector.stop()
        self.world.stop()

    def joinSuccessful(self, agent, playerId):
        '''
        Called when a join request succeeds.
        '''
        player = self.world.getPlayer(playerId)
        assert player.agent is None
        assert agent.player is None

        player.agent = agent
        agent.setPlayer(player)

    def playerRemoved(self, player, playerId):
        '''
        Called when a player is removed from the game.
        '''
        if player.agent:
            player.agent.setPlayer(None)


class LocalGame(Game):
    def __init__(
            self, layoutDatabase, halfMapWidth, mapHeight, blockRatio=0.5,
            duration=0, maxPerTeam=8, maxTotal=40, authTagManager=None,
            onceOnly=False, ais=True, saveReplay=False,
            gamePrefix=None, replayPath=None):

        self.maxPerTeam = maxPerTeam
        self.maxTotalPlayers = min(2 * maxPerTeam, maxTotal)
        self.authTagManager = authTagManager

        self.agents = set()
        self.agentInfos = {}
        self.layoutDatabase = layoutDatabase

        self.pendingZombieNotices = []

        world = ServerUniverse(
            self, layoutDatabase, authTagManager=authTagManager,
            duration=duration, onceOnly=onceOnly)

        self.idManager = world.idManager

        self.gameRecorder = GameRecorder(
            self, world, saveReplay=saveReplay, gamePrefix=gamePrefix,
            replayPath=replayPath)

        Game.__init__(self, world, ais=ais)

        layout = layoutDatabase.generateRandomMapLayout(
            halfMapWidth, mapHeight, blockRatio=blockRatio)
        world.setLayout(layout)
        world.onServerTickComplete.addListener(self.worldTickDone)
        self.gameRecorder.start()

        self.updateDelayLoop = WeakLoopingCall(self, 'updateDelays')
        self.updateDelayLoop.start(10, False)

        self.achievementManager = AchievementManager(self)
        self.achievementManager.start()

    def updateDelays(self):
        for info in self.agentInfos.itervalues():
            info.updateDelays()

    def addAgent(self, agent):
        info = AgentInfo(self, agent)
        self.agentInfos[agent] = info
        self.agents.add(agent)

    def detachAgent(self, agent):
        self.agents.remove(agent)
        if agent.player:
            self.kickPlayer(agent.player.id)
        del self.agentInfos[agent]

    def kickPlayer(self, playerId):
        '''
        Removes the player with the specified ID from the game.
        '''
        self.sendServerCommand(RemovePlayerMsg(playerId))

    def agentRequest(self, agent, msg):
        '''
        Some messages need to be delayed until the correct time comes.
        '''
        msg.tracePoint(self, 'agentRequest')
        if agent not in self.agents:
            log.warning('LocalGame got message %s from unconnected agent', msg)
            return
        info = self.agentInfos[agent]
        info.requestFromAgent(msg)

    def dispatchDelayedRequest(self, agent, msg):
        '''
        Called by an AgentInfo when the correct time comes for a request to be
        dispatched.
        '''
        msg.tracePoint(self, 'dispatchDelayedRequest')
        msg.serverApply(self, agent)

    def setPlayerLimits(self, maxPerTeam, maxTotal=40):
        '''
        Changes the player limits in the current game. Note that this does not
        affect players who are already in the game.

        @param maxPerTeam: Maximum number of players per team at once
        @param maxTotal: Maximum number of players in the game at once
        '''
        self.maxPerTeam = maxPerTeam
        self.maxTotalPlayers = min(2 * maxPerTeam, maxTotal)

    def getGameState(self):
        '''
        Returns the current state of the game as a string.

        @returns: The text id of the current game state
        '''
        currentGameState = self.world.state
        if currentGameState is not None:
            return currentGameState.id
        return 'Unknown'

    def stop(self):
        super(LocalGame, self).stop()
        self.gameRecorder.stop()
        self.achievementManager.stop()
        self.updateDelayLoop.stop()
        self.idManager.stop()

    def noticedZombieHit(self, info, msg):
        self.pendingZombieNotices.append((info, msg))

    def worldTickDone(self):
        '''
        Called when the universe has ticked.
        '''
        while self.pendingZombieNotices:
            info, msg = self.pendingZombieNotices.pop(0)
            info.passRequestToGame(msg, delay=info.currentDelay)

        self.checkCollisionsWithCollectables()
        for info in self.agentInfos.itervalues():
            info.tick()

    def checkCollisionsWithCollectables(self):
        '''
        When a player runs into a collectable unit (e.g. a star or the
        trosball), we pay attention not to where we think the collectable units
        are, but where the player's client thinks they are. To do this, we need
        to project the collectable units back in time based on the player's
        current delay.
        '''
        greatestDelay = 0
        for info in self.agentInfos.itervalues():
            if not info.player or info.player.dead:
                continue
            greatestDelay = max(greatestDelay, info.currentDelay)

            for unit in self.world.getCollectableUnits():
                if unit.checkCollision(info.player, info.currentDelay):
                    unit.collidedWithPlayer(info.player)

        for unit in self.world.getCollectableUnits():
            unit.clearOldHistory(greatestDelay)

    def joinSuccessful(self, agent, playerId):
        super(LocalGame, self).joinSuccessful(agent, playerId)
        self.agentInfos[agent].givePlayer(self.world.getPlayer(playerId))

    def playerRemoved(self, player, playerId):
        super(LocalGame, self).playerRemoved(player, playerId)
        info = self.agentInfos[player.agent]
        info.takePlayer()

    def sendServerCommand(self, msg):
        '''
        Sends a command to the universe and all attached agents. Typically
        called by message classes in serverApply().
        '''
        msg.tracePoint(self, 'sendServerCommand')
        self.world.consumeMsg(msg)
        self.gameRecorder.consumeMsg(msg)
        self.onServerCommand(msg)

    def sendResync(
            self, playerId,
            reason='Your computer was out of sync with the server!'):
        '''
        Resyncs the position of the player with the given id. Typically called
        by message classes in serverApply().
        '''
        player = self.world.getPlayer(playerId)
        player.sendResync(reason)


class AgentInfo(object):
    '''
    Used by a local game to keep track of information about a connected Agent.
    '''

    def __init__(self, game, agent):
        self.game = game
        self.agent = agent
        self.world = game.world
        self.player = None

        self.currentDelay = INITIAL_ASSUMED_LATENCY
        self.recentDelay = -1
        self.idealDelay = int(0.5 * INITIAL_ASSUMED_LATENCY)
        self.requestQueue = []

    def requestFromAgent(self, msg):
        '''
        When we receive a request from an agent, we first check to see if this
        request is supposed to occur at a specific time. If so, we delay the
        message until that time.
        '''
        msg.tracePoint(self, 'requestFromAgent')
        if msg.timestampedPlayerRequest:
            measuredDelay = self.world.lastTickId - msg.tickId
            if measuredDelay > LAG_THRESHOLD:
                adjustedDelay = measuredDelay + LAG_BUFFER
            else:
                adjustedDelay = measuredDelay
            if __debug__ and globaldebug.enabled:
                adjustedDelay = max(adjustedDelay, globaldebug.forceDelay)
            self.recentDelay = max(self.recentDelay, adjustedDelay)

            if measuredDelay > self.currentDelay:
                self.currentDelay = adjustedDelay
                if isinstance(msg, ResyncAcknowledgedMsg):
                    self.passRequestToGame(
                        msg, delay=(adjustedDelay - measuredDelay))
                else:
                    self.player.sendResync(
                        reason='Your latency just spiked to %.2f s RTT' % (
                            measuredDelay * TICK_PERIOD))
                self.agent.messageToAgent(DelayUpdatedMsg(self.currentDelay))
            else:
                self.passRequestToGame(
                    msg, delay=(self.currentDelay - measuredDelay))
        else:
            self.passRequestToGame(msg)

    def passRequestToGame(self, msg, delay=0):
        msg.tracePoint(self, 'passRequestToGame')
        if delay == 0:
            self.game.dispatchDelayedRequest(self.agent, msg)
        else:
            while len(self.requestQueue) < delay:
                self.requestQueue.append([])
            self.requestQueue[delay - 1].append(msg)

    def updateDelays(self):
        self.idealDelay, self.recentDelay = self.recentDelay, -1

    def tick(self):
        if self.player:
            if __debug__ and globaldebug.enabled:
                if globaldebug.localPlayerId == self.player.id:
                    globaldebug.localPlayerDelay = self.currentDelay
            static = self.player.isStatic()
        else:
            static = False

        if self.idealDelay < 0 and self.recentDelay < 0:
            # We have not seen any recent timed requests so we don't know how
            # much this agent is really lagging by.
            return

        idealDelay = max(self.idealDelay, self.recentDelay, 0)
        if static:
            changed = False
            while idealDelay < self.currentDelay:
                if self.requestQueue and self.requestQueue[0]:
                    break
                log.debug(
                    'Catching up one frame of player lag (%s)',
                    self.player.nick if self.player else '-')
                if self.requestQueue:
                    self.requestQueue.pop(0)
                changed = True
                self.currentDelay -= 1
            if changed:
                self.agent.messageToAgent(DelayUpdatedMsg(self.currentDelay))

        if self.requestQueue:
            for msg in self.requestQueue.pop(0):
                self.passRequestToGame(msg)

    def givePlayer(self, player):
        '''
        Called when this agent is being given control of the specified player.
        '''
        assert self.player is None
        self.player = player
        player.onZombieHit.addListener(self.zombieHit)

    def takePlayer(self):
        if not self.player:
            return
        self.player.onZombieHit.removeListener(self.zombieHit)
        self.player = None

    def zombieHit(self, killer, shot, deathType):
        '''
        Queue a fake message to remind us when the universe gets to the state
        it will be in when the client notices that it's been hit.
        '''
        self.game.noticedZombieHit(self, PlayerNoticedZombieHitMsg(
            self.player.id, killer.id, deathType))


class RemoteGame(Game, Node):
    '''
    Represents a game that is running on a remote server. This game object
    maintains the current state of the Universe based on messages from the
    server.
    '''

    def __init__(self, layoutDatabase, *args, **kwargs):
        world = Universe(layoutDatabase)
        super(RemoteGame, self).__init__(world, None, *args, **kwargs)
        self.agentIds = {}
        self.agentById = {}

    def connected(self, settings):
        self.world.restoreEverything(settings)

    @defer.inlineCallbacks
    def addAgent(self, agent):
        assert agent not in self.agentIds
        agentId = yield self.hub.connectNewAgent()
        self.agentIds[agent] = agentId
        self.agentById[agentId] = agent

    def detachAgent(self, agent):
        if agent in self.agentIds:
            self.hub.disconnectAgent(self.agentIds[agent])

    def agentRequest(self, agent, msg):
        '''
        Called by an agent with an AgentRequest message that it wants to
        request.
        '''
        msg.tracePoint(self, 'agentRequest')
        self.hub.sendRequestToGame(self.agentIds[agent], msg)

    def gotServerCommand(self, msg):
        '''
        Called when a server command is received from the network.
        '''
        msg.tracePoint(self, 'gotServerCommand')
        self.world.consumeMsg(msg)
        self.onServerCommand(msg)

    def gotMessageToAgent(self, agentId, msg):
        msg.tracePoint(self, 'gotMessageToAgent')
        agent = self.agentById[agentId]
        agent.messageToAgent(msg)

    def agentDisconnected(self, agentId):
        agent = self.agentById[agentId]
        del self.agentIds[agent]
        del self.agentById[agentId]
        agent.gotServerCommand(ConnectionLostMsg())


class AIInjector(object):
    def __init__(self, game, playerCount=6):
        self.game = game
        self.playerCount = playerCount
        self.loop = WeakLoopingCall(self, 'tick')
        self.agents = []
        self.newAgents = set()
        self.aiNames = listAIs(playableOnly=True)

    def start(self):
        self.loop.start(3, False)

    def stop(self):
        if self.loop.running:
            self.loop.stop()
        self._stopAllAgents()

    @defer.inlineCallbacks
    def tick(self):
        self._graduateNewAgents()

        if self.game.world.state != gameStates.Lobby:
            self._stopAllAgents()
        else:
            yield self._adjustAgentsToTarget()

    def _graduateNewAgents(self):
        for agent in list(self.newAgents):
            if agent.player is not None:
                self.agents.append(agent)
                self.newAgents.remove(agent)

    def _stopAllAgents(self):
        if len(self.agents) != 0:
            log.info('AIInjector: Stopping all agents')
        for agent in self.agents:
            agent.stop()
            self.game.detachAgent(agent)
        self.agents = []

    @defer.inlineCallbacks
    def _adjustAgentsToTarget(self):
        worldPlayers = len(self.game.world.players)
        newAgents = len(self.newAgents)
        if self.playerCount > worldPlayers + newAgents:
            yield self._addAgents(self.playerCount - worldPlayers - newAgents)
        else:
            self._removeAgents(worldPlayers + newAgents - self.playerCount)

    @defer.inlineCallbacks
    def _addAgents(self, count):
        log.info('AIInjector: Adding %d agents', count)
        for i in xrange(count):
            agent = yield self.game.addAI(random.choice(self.aiNames))
            self.newAgents.add(agent)

    def _removeAgents(self, count):
        if count != 0:
            log.info('AIInjector: Removing %d agents', count)
        for i in xrange(count):
            if len(self.agents) == 0:
                break
            agent = self.agents.pop(0)
            agent.stop()
            self.game.detachAgent(agent)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

import pygame

from framework import Element
from elements import TextElement
from trosnoth.utils.event import Event
from trosnoth.gui.common import AttachedPoint, Location, ScaledSize

log = logging.getLogger('checkbox')

class CheckBox(Element):
    def __init__(self, app, pos, text, font, colour, initValue=False,
                 hotkey=None, style='circle', fillColour=None):
        super(CheckBox, self).__init__(app)

        self.pos = pos
        self.font = font

        if hasattr(pos, 'apply'):
            self.text = TextElement(app, ' ' + text, font,
                    Location(AttachedPoint(ScaledSize(self._getBoxSize()[0] / 5,
                    2), self._getBoxRect, 'midright'), 'midleft'), colour)
        else:
            self.text = TextElement(app, ' ' + text, font, Location((pos[0] +
                    self._getBoxSize()[0], pos[1] - self._getBoxSize()[0] / 10),
                    'topleft'), colour)

        self.value = initValue
        self.colour = colour
        if fillColour is None:
            self.fillColour = tuple((256*3+i)/4 for i in colour)
        else:
            self.fillColour = fillColour
        self.hotkey = hotkey
        self.style = style
        self.onValueChanged = Event()

    def _getRect(self):
        return self._getBoxRect().union(self.text._getRect())

    def _getBorderWidth(self):
        return max(1, self._getBoxSize()[0] / 8)

    def _getBoxSize(self):
        return (int(self.font.getHeight(self.app) / 1.5),
                int(self.font.getHeight(self.app) / 1.5))

    def _getBoxRect(self):
        if hasattr(self.pos, 'apply'):
            box = pygame.rect.Rect((0,0), self._getBoxSize())
            self.pos.apply(self.app, box)
        else:
            box = pygame.rect.Rect(self.pos, self._getBoxSize())
        return box

    def draw(self, surface):
        box = self._getBoxRect()
        if self.value:
            if self.style == 'fill':
                pygame.draw.rect(surface, self.fillColour, box, 0)
            elif self.style == 'cross':
                pygame.draw.line(surface, self.fillColour, box.topright,
                        box.bottomleft, self._getBorderWidth())
                pygame.draw.line(surface, self.fillColour, box.topleft,
                        box.bottomright, self._getBorderWidth())
            elif self.style == 'circle':
                pygame.draw.circle(surface, self.fillColour, box.center,
                        box.width / 2 - 2)
        pygame.draw.rect(surface, self.colour, box, self._getBorderWidth())

        self.text.draw(surface)

    def setValue(self, val):
        if val != self.value:
            self.value = val
            self.onValueChanged.execute(self)

    def getValue(self):
        return self.value

    def processEvent(self, event):
        box = self._getBoxRect()
        # Active events.
        if (event.type == pygame.MOUSEBUTTONDOWN and event.button == 1 and
                box.collidepoint(event.pos)) or (event.type == pygame.KEYDOWN
                and event.key == self.hotkey):
            self.setValue(not self.value)
            self.onValueChanged.execute(self)
        else:
            return event
        return None



import logging
import random
import datetime
import pygame

from trosnoth.const import MAP_TO_SCREEN_SCALE, COLLECTABLE_STAR_LIFETIME
from trosnoth.gui.framework.basics import (
    AngledImageCollection, Animation, SingleImage,
)
from trosnoth.model.player import Player
from trosnoth.trosnothgui.common import setAlpha
from trosnoth.trosnothgui.ingame.nametag import (
    NameTag, StarTally, HealthBar, CountDown,
)

log = logging.getLogger(__name__)


class UnitSprite(pygame.sprite.Sprite):
    def __init__(self, app, worldGUI, unit):
        super(UnitSprite, self).__init__()
        self.app = app
        self.worldGUI = worldGUI
        self.unit = unit

    @property
    def pos(self):
        return self.unit.tweenPos(self.worldGUI.tweenFraction)


class ShotSprite(UnitSprite):
    def __init__(self, app, worldGUI, shot):
        super(ShotSprite, self).__init__(app, worldGUI, shot)
        self.shot = shot
        self.shotAnimation = app.theme.sprites.shotAnimation(shot.team)
        # Need a starting one:
        self.image = self.shotAnimation.getImage()
        self.rect = self.image.get_rect()

    def update(self):
        self.image = self.shotAnimation.getImage()


class SingleAnimationSprite(pygame.sprite.Sprite):
    def __init__(self, app, pos):
        super(SingleAnimationSprite, self).__init__()
        self.app = app
        self.pos = pos
        self.animation = self.getAnimation()
        self.image = self.animation.getImage()
        self.rect = self.image.get_rect()

    def getAnimation(self):
        raise NotImplementedError('getAnimation')

    def update(self):
        self.image = self.animation.getImage()

    def isDead(self):
        return self.animation.isComplete()


class ExplosionSprite(SingleAnimationSprite):
    def getAnimation(self):
        return self.app.theme.sprites.explosion()


class ShoxwaveExplosionSprite(SingleAnimationSprite):
    def getAnimation(self):
        return self.app.theme.sprites.shoxwaveExplosion()


class TrosballExplosionSprite(SingleAnimationSprite):
    def getAnimation(self):
        return self.app.theme.sprites.trosballExplosion()


class GrenadeSprite(UnitSprite):
    def __init__(self, app, worldGUI, grenade):
        super(GrenadeSprite, self).__init__(app, worldGUI, grenade)
        self.grenade = grenade
        self.image = app.theme.sprites.teamGrenade(grenade.player.team)
        self.rect = self.image.get_rect()


class CollectableStarSprite(UnitSprite):
    def __init__(self, app, worldGUI, star):
        super(CollectableStarSprite, self).__init__(app, worldGUI, star)
        self.star = star
        self.image = app.theme.sprites.star
        self.alphaImage = self.image.copy()
        self.rect = self.image.get_rect()

    def update(self):
        time = self.worldGUI.universe.timer.getElapsedTime()
        if self.star.creationTime + COLLECTABLE_STAR_LIFETIME < time + 1:
            alpha = random.randint(32, 192)
            self.image = self.app.theme.sprites.star.copy()
            setAlpha(self.image, alpha, alphaSurface=self.alphaImage)


class TrosballSprite(pygame.sprite.Sprite):
    def __init__(self, app, worldGUI, world):
        super(TrosballSprite, self).__init__()
        self.app = app
        self.worldGUI = worldGUI
        self.world = world
        self.animation = app.theme.sprites.trosballAnimation()
        self.warningAnimation = app.theme.sprites.trosballWarningAnimation()
        # Need a starting one:
        self.image = self.animation.getImage()
        self.rect = self.image.get_rect()

    def update(self):
        self.image = self.animation.getImage()
        if self.world.trosballPlayer is not None:
            trosballExplodeTime = (
                self.world.playerGotTrosballTime +
                self.world.physics.trosballExplodeTime)
            if trosballExplodeTime - self.world.timer.getElapsedTime() < 2:
                self.image = self.warningAnimation.getImage()
        center = self.rect.center
        self.rect = self.image.get_rect()
        self.rect.center = center

    @property
    def pos(self):
        if self.world.trosball:
            x, y = self.world.trosball.tweenPos(self.worldGUI.tweenFraction)
        elif self.world.trosballPlayer:
            p = self.world.trosballPlayer
            if p.id == self.worldGUI.localPlayerId:
                p = self.worldGUI.localPlayerSprite
            x, y = p.tweenPos(self.worldGUI.tweenFraction)
            x += 5 if p._faceRight else -5
        else:
            layout = self.world.map.layout
            x, y = layout.centreX, layout.centreY
        return (x, y)


class PlayerSprite(UnitSprite):
    # These parameters are used to create a canvas for the player sprite object
    canvasSize = (
        int(33 * MAP_TO_SCREEN_SCALE + 0.5),
        int(39 * MAP_TO_SCREEN_SCALE + 0.5))
    liveOffset = 3
    ghostOffset = 0

    def __init__(self, app, worldGUI, player, greyed=False, timer=None):
        super(PlayerSprite, self).__init__(app, worldGUI, player)
        if timer is None:
            timer = player.world.timer.getElapsedTime
        self.spriteTeam = player.team
        self.player = player
        self.nametag = NameTag(app, player.nick)
        self.countdown = CountDown(app, self.player)
        self._oldName = player.nick
        self.starTally = StarTally(app, 0)
        self.greyed = greyed
        self.healthBar = HealthBar(
            app,
            badColour=self.app.theme.colours.badHealth,
            fairColour=self.app.theme.colours.fairHealth,
            goodColour=self.app.theme.colours.goodHealth)
        self.shieldBar = HealthBar(
            app,
            badColour=self.app.theme.colours.badShield,
            fairColour=self.app.theme.colours.fairShield,
            goodColour=self.app.theme.colours.goodShield)

        sprites = app.theme.sprites
        self.greyVersion = None
        if greyed:
            sprites = sprites.greyed
        self.sprites = sprites
        self.gunImages = AngledImageCollection(
            self.getAngleFacing, *sprites.gunImages(self.player.team))
        self.machineGunImages = AngledImageCollection(
            self.getAngleFacing, *sprites.machineGunImages(self.player.team))
        self.ricoGunImages = AngledImageCollection(
            self.getAngleFacing, *sprites.ricoGunImages(self.player.team))
        self.shoxGunImages = SingleImage(sprites.shoxGun3)

        head = sprites.playerHead(self.player.team, self.player.bot)
        self.runningAnimation = [
            Animation(0.1, timer, *sprites.runningLegs),
            sprites.playerBody,
            head,
        ]
        self.ghostAnimation = sprites.ghostAnimation(self.player.team)
        self.blocker = [
            Animation(
                0.1, timer,
                sprites.blocker1, sprites.blocker1, sprites.blocker2),
            head,
        ]
        self.bomber = [
            head,
            Animation(0.08, timer, sprites.bomber1, sprites.bomber2),
        ]

        self.reversingAnimation = [
            sprites.playerBody,
            Animation(0.1, timer, *sprites.backwardsLegs),
            head,
        ]

        self.turretAnimation = [
            sprites.turretBase,
            sprites.playerBody,
            head,
        ]

        self.standingAnimation = [
            sprites.playerStanding,
            sprites.playerBody,
            head,
        ]

        self.jumpingAnimation = [
            sprites.playerJumping,
            sprites.playerBody,
            head,
        ]
        self.holdingAnimation = [
            sprites.playerBody,
            sprites.playerHolding(self.player.team),
            head,
        ]
        self.fallingAnimation = self.jumpingAnimation
        self.shieldAnimation = Animation(0.15, timer, *sprites.shieldImages)
        self.jammingHat = sprites.jammingHat()

        if self.app.displaySettings.perPixelAlpha:
            flags = pygame.SRCALPHA
            self.alphaImage = pygame.Surface(self.canvasSize, flags)
        else:
            flags = 0
            self.alphaImage = None

        self.image = pygame.Surface(self.canvasSize, flags)
        self.rect = self.image.get_rect()

        # This probably shouldn't be done here.
        _t = datetime.date.today()
        self.is_christmas = _t.day in (24, 25, 26) and _t.month == 12

    def getAngleFacing(self):
        return self.player.angleFacing

    @property
    def angleFacing(self):
        return self.player.angleFacing

    def __getattr__(self, attr):
        '''
        Proxy attributes through to the underlying player class.
        '''
        return getattr(self.player, attr)

    def update(self):
        if self.player.nick != self._oldName:
            self._oldName = self.player.nick
            self.nametag = NameTag(self.app, self.player.nick)

        self.setImage()

    def _isSlow(self):
        # Consider horizontal movement of player.
        xMotion = self.player.getXKeyMotion()
        if xMotion < 0:
            return self.player._faceRight
        if xMotion > 0:
            return not self.player._faceRight
        return False

    def setImage(self):
        if not self.app.displaySettings.perPixelAlpha:
            self.image.set_alpha(None)

        self.image.fill((127, 127, 127, 0))
        self.image.set_colorkey((127, 127, 127))

        if self.player.resyncing and not self.greyed:
            if self.greyVersion is None:
                self.greyVersion = PlayerSprite(
                    self.app, self.worldGUI, self.player, greyed=True)
            self.greyVersion.setImage()
            self.image.blit(self.greyVersion.image, (0, 0))
            return

        flip = None
        offset = self.liveOffset
        if self.player.dead:
            blitImages = self.ghostAnimation
            offset = self.ghostOffset
        elif self.player.turret:
            blitImages = self.turretAnimation
        elif self.player.bomber:
            if self.player.upgradeGauge < 0.8:
                blitImages = self.bomber
            else:
                blitImages = self.blocker
        elif self.player.isAttachedToWall():
            blitImages = self.holdingAnimation
            if self.player.isAttachedToWall() == 'right':
                flip = False
            else:
                flip = True
        elif self.player.isOnGround():
            if self.player.getXKeyMotion() == 0:
                blitImages = self.standingAnimation
            elif self._isSlow():
                blitImages = self.reversingAnimation
            else:
                blitImages = self.runningAnimation
        else:
            if self.player.yVel > 0:
                blitImages = self.fallingAnimation
            else:
                blitImages = self.jumpingAnimation

        # Put the pieces together:
        for element in blitImages:
            self.image.blit(element.getImage(), (offset, 0))
        if not (
                self.player.dead or self.player.isAttachedToWall() or
                self.player.bomber):
            if self.player.machineGunner:
                weapon = self.machineGunImages
            elif self.player.hasRicochet:
                weapon = self.ricoGunImages
            elif self.player.shoxwave:
                weapon = self.shoxGunImages
            else:
                weapon = self.gunImages
            self.image.blit(weapon.getImage(), (offset, 0))
        if self.player.ninja:
            self.image.blit(
                self.sprites.ninjaHead.getImage(), (offset, 0))
        elif self.player.disruptive:
            self.image.blit(self.jammingHat.getImage(), (offset, 0))
        if self.player.hasElephant() and not self.player.dead:
            self.image.blit(
                self.sprites.elephant.getImage(), (offset, 0))
        if (not self.player.dead and not self.player.phaseshift and not
                self.player.ninja and self.is_christmas and not
                self.player.hasElephant()):
            self.image.blit(
                self.sprites.christmasHat.getImage(), (offset, 0))
        if not self.player._faceRight and flip is None or flip:
            self.image = pygame.transform.flip(self.image, True, False)
        if self.player.hasVisibleShield():
            img = self.shieldAnimation.getImage()
            if not (img.get_flags() & pygame.SRCALPHA):
                # The shield animation already uses per-pixel alphas so if they
                # are enabled we don't need per-surface alphas.
                img.set_alpha(128)
            self.image.blit(img, (offset, 0))

        # Flicker the sprite between different levels of transparency
        if self.player.phaseshift and self._canSeePhaseShift():
            self.setImageAlpha(random.randint(30, 150))
        elif self.player.dead:
            self.setImageAlpha(128)
        elif self.player.isInvulnerable():
            self.setImageAlpha(random.randint(30, 150))
        elif self.player.invisible:
            replay = self.worldGUI.gameViewer.replay
            if replay or self.player.isFriendsWith(self.getShownPlayer()):
                self.setImageAlpha(80)
            else:
                self.setImageAlpha(0)
        else:
            self.setImageAlpha(255)

    def setImageAlpha(self, alpha):
        setAlpha(self.image, alpha, alphaSurface=self.alphaImage)

    def getShownPlayer(self):
        return self.worldGUI.gameViewer.viewManager.target

    def _canSeePhaseShift(self):
        if self.worldGUI.gameViewer.replay:
            return True
        target = self.getShownPlayer()
        if not isinstance(target, Player):
            return False
        return self.player.isFriendsWith(target)



import logging

from twisted.internet.protocol import ClientCreator
from twisted.internet import reactor, defer
from twisted.protocols import amp

from trosnoth.gui.framework import framework, table, elements, scrollableCanvas
from trosnoth.gui.framework.prompt import intValidator, PasswordBox
from trosnoth.gui.framework.elements import TextButton, TextElement
from trosnoth.gui.framework.checkbox import CheckBox
from trosnoth.gui.common import (ScaledScreenAttachedPoint, Location,
        ScaledScalar, ScaledSize, ScaledArea, Area, AttachedPoint, Size, Region,
        ScaledPoint)
from trosnoth.gui.framework.tabContainer import TabContainer, TabSize
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.framework.dialogbox import OkBox
from trosnoth.network import authcommands
from trosnoth.network.lobby import authenticate, encryptPassword
from trosnoth.trosnothgui.common import button
from trosnoth.trosnothgui.pregame.authServerLoginBox import PasswordGUI
from trosnoth.trosnothgui.pregame.playscreen import PlayAuthScreen

log = logging.getLogger('serverselection')

HOSTNAME_COLUMN = 0
PORT_COLUMN = 1
CONNECT_COLUMN = 2
ACCOUNT_COLUMN = 3
MOVE_UP_COLUMN = 4
MOVE_DOWN_COLUMN = 5
DELETE_COLUMN = 6

class AccountSettingsScreen(framework.CompoundElement):
    passwordGUIFactory = PasswordGUI

    def __init__(self, app, host, port, onClose):
        super(AccountSettingsScreen, self).__init__(app)
        self.onClose = onClose
        self.host = host
        self.port = port

        area = ScaledArea(50,140,924, 570)
        alpha = 192 if app.displaySettings.alphaOverlays else 255
        font = app.screenManager.fonts.bigMenuFont
        self.tabContainer = TabContainer(self.app, area, font,
                app.theme.colours.playTabBorder)
        self.background = elements.SolidRect(self.app,
                app.theme.colours.playMenu, alpha, Area(AttachedPoint((0,0),
                self.tabContainer._getTabRect), TabSize(self.tabContainer)))

        self.passwordTab = ChangePasswordTab(app, host, onClose=self.close,
                onSave=self.save)
        self.tabContainer.addTab(self.passwordTab)
        self.passwordGetter = self.passwordGUIFactory(self.app)

        self.elements = [self.background, self.tabContainer]

        self.protocol = None
        d = ClientCreator(reactor, amp.AMP).connectTCP(host, port)
        d.addCallbacks(self.connectionEstablished, self.connectionFailed)

    @defer.inlineCallbacks
    def connectionEstablished(self, protocol):
        self.protocol = protocol
        try:
            yield authenticate(protocol, self.host, self.passwordGetter)
        except:
            OkBox(self.app, ScaledSize(450, 150), 'Trosnoth',
                    'Unable to authenticate with server').show()
            self.close()
            return

        try:
            result = yield self.protocol.callRemote(
                    authcommands.GetSupportedSettings)
            tabs = set(result['result'])
            self.showTabs(tabs)
        except:
            log.error('Error calling GetSupportedSettings', exc_info=True)
            OkBox(self.app, ScaledSize(450, 150), 'Trosnoth',
                    'Error communicating with server').show()
            self.close()
            return

    def showTabs(self, tabs):
        tabs = set(tabs)

    def connectionFailed(self, reason):
        box = OkBox(self.app, ScaledSize(450, 150), 'Trosnoth',
                'Could not connect to server')
        box.show()
        self.close()

    def close(self):
        if self.protocol and self.protocol.transport.connected:
            self.protocol.transport.loseConnection()
        self.onClose()

    @defer.inlineCallbacks
    def save(self):
        try:
            failReason = yield self.passwordTab.save(self.protocol)
        except:
            log.error('Error saving account settings', exc_info=True)
            failReason = 'Error saving settings to server'

        if failReason:
            OkBox(self.app, ScaledSize(450, 150), 'Trosnoth', failReason).show()
        else:
            self.close()


class ServerSelectionScreen(framework.CompoundElement):
    connectScreenFactory = PlayAuthScreen
    accountScreenFactory = AccountSettingsScreen

    def __init__(self, app, onClose):
        super(ServerSelectionScreen, self).__init__(app)
        self.onClose = onClose

        area = ScaledArea(50,140,924, 570)
        if app.displaySettings.alphaOverlays:
            alpha = 192
        else:
            alpha = 255
        font = app.screenManager.fonts.bigMenuFont
        self.tabContainer = TabContainer(self.app, area, font,
                app.theme.colours.playTabBorder)
        self.background = elements.SolidRect(self.app,
                app.theme.colours.playMenu, alpha, Area(AttachedPoint((0,0),
                self.tabContainer._getTabRect), TabSize(self.tabContainer)))

        self.tab = ServerSelectionTab(app, onClose=onClose, onJoin=self.join,
                onAccountSettings=self.accountSettings)
        self.tabContainer.addTab(self.tab)

        self.setElements()

    def join(self, servers):
        playAuthScreen = self.connectScreenFactory(self.app,
                onSucceed=self.onClose, onFail=self.joinFailed)
        self.elements = [playAuthScreen]
        playAuthScreen.begin(servers, canHost=False)

    def joinFailed(self):
        self.setElements()

    def reload(self):
        self.tab.reload()
        self.setElements()

    def setElements(self):
        self.elements = [self.background, self.tabContainer]

    def accountSettings(self, host, port):
        accountScreen = self.accountScreenFactory(self.app, host, port,
                onClose=self.reload)
        self.elements = [accountScreen]

class ServerSelectionTab(Tab, framework.CompoundElement):
    def __init__(self, app, onClose, onJoin, onAccountSettings):
        super(ServerSelectionTab, self).__init__(app, 'Server Settings')
        self.onClose = onClose
        self.onJoin = onJoin
        self.onAccountSettings = onAccountSettings
        self.setupButtons()

    def setupButtons(self):
        lanButtonPos = Location(ScaledScreenAttachedPoint(ScaledPoint(100, 550),
                'topleft'))
        self.lanButton = CheckBox(self.app, lanButtonPos, 'LAN Discovery',
                self.app.screenManager.fonts.serverSelectionCheckboxesFont,
                self.app.theme.colours.mainMenuColour,
                initValue=self.app.connectionSettings.lanGames, style='circle',
                fillColour=self.app.theme.colours.mainMenuColour)
        self.lanButton.onValueChanged.addListener(lambda sender: self.save())

        lanSearchButtonPos = Location(ScaledScreenAttachedPoint(
                ScaledPoint(380, 550), 'topleft'))
        self.lanSearchButton = TextButton(self.app, lanSearchButtonPos,
                'Search LAN for Game',
                self.app.screenManager.fonts.serverSelectionCheckboxesFont,
                self.app.theme.colours.mainMenuHighlight,
                self.app.theme.colours.white,
                onClick=lambda sender:self.joinLan())

        createButtonPos = Location(ScaledScreenAttachedPoint(
                ScaledPoint(100, 600), 'topleft'))
        self.createButton = CheckBox(self.app, createButtonPos,
                'Create games on remote servers',
                self.app.screenManager.fonts.serverSelectionCheckboxesFont,
                self.app.theme.colours.mainMenuColour,
                initValue=self.app.connectionSettings.createGames,
                style='circle',
                fillColour=self.app.theme.colours.mainMenuColour)
        self.createButton.onValueChanged.addListener(lambda sender: self.save())

        otherButtonPos = Location(ScaledScreenAttachedPoint(
                ScaledPoint(100, 650), 'topleft'))
        self.otherButton = CheckBox(self.app, otherButtonPos,
                'Ask servers about other games',
                self.app.screenManager.fonts.serverSelectionCheckboxesFont,
                self.app.theme.colours.mainMenuColour,
                initValue=self.app.connectionSettings.otherGames,
                style='circle',
                fillColour=self.app.theme.colours.mainMenuColour)
        self.otherButton.onValueChanged.addListener(lambda sender: self.save())

        newButtonPos = Location(ScaledScreenAttachedPoint(
                ScaledPoint(850, 285), 'topleft'))
        self.newButton = TextButton(self.app, newButtonPos, '+ Server',
                self.app.screenManager.fonts.serverSelectionCheckboxesFont,
                self.app.theme.colours.serverSelectionNewItem,
                self.app.theme.colours.white,
                onClick=lambda sender:self.newServer())

        self.closeButton = elements.TextButton(self.app,
            Location(ScaledScreenAttachedPoint(
                    ScaledSize(-70, 650), 'topright'), 'topright'),
            'done',
            self.app.screenManager.fonts.bigMenuFont,
            self.app.theme.colours.mainMenuHighlight,
            self.app.theme.colours.white,
            onClick=lambda sender:self.onClose(),
        )

    def setupTable(self):
        position = Location(ScaledScreenAttachedPoint(
                ScaledPoint(100, 200), 'topleft'), 'topleft')
        # position is merely temporary
        self.serverTable = table.Table(self.app, position)
        hostColumn = table.TextBoxColumn(self.app, self.serverTable,
                ScaledScalar(290))
        portColumn = table.TextBoxColumn(self.app, self.serverTable,
                ScaledScalar(70))
        connectButtonColumn = table.TextButtonColumn(self.app, self.serverTable,
                ScaledScalar(80))
        accountButtonColumn = table.TextButtonColumn(self.app, self.serverTable,
                ScaledScalar(100))
        upButtonColumn = table.TextButtonColumn(self.app, self.serverTable,
                ScaledScalar(40))
        downButtonColumn = table.TextButtonColumn(self.app, self.serverTable,
                ScaledScalar(40))
        deleteButtonColumn = table.TextButtonColumn(self.app, self.serverTable,
                ScaledScalar(40))

        self.serverTable.addColumns([
            hostColumn,
            portColumn,
            connectButtonColumn,
            accountButtonColumn,
            upButtonColumn,
            downButtonColumn,
            deleteButtonColumn,
        ])

        self.serverTable.setBorderWidth(4)
        self.serverTable.setBorderColour((0, 0, 0))

        with self.serverTable.style as s:
            s.backColour = (255, 255, 255)
            s.foreColour = self.app.theme.colours.mainMenuColour
            s.font = self.app.screenManager.fonts.smallMenuFont
            s.padding = (4, 1)
            s.hasShadow = True
            s.shadowColour = (0, 0, 0)

        hostColumn.style.textAlign = 'midright'
        portColumn.style.textAlign = 'center'

        for column in (connectButtonColumn, accountButtonColumn, upButtonColumn,
                downButtonColumn):
            with column.style as s:
                s.textAlign = 'center'
                s.foreColour = self.app.theme.colours.mainMenuHighlight
                s.hoverColour = self.app.theme.colours.black

        with deleteButtonColumn.style as s:
            s.foreColour = self.app.theme.colours.red
            s.backColour = self.app.theme.colours.black
            s.hoverColour = self.app.theme.colours.white
            s.textAlign = 'center'

        self.serverTable.setDefaultHeight(ScaledScalar(40))

        # Title row
        row = self.serverTable.addRow()
        row[HOSTNAME_COLUMN].setText('Host Name')
        row[HOSTNAME_COLUMN].style.foreColour = self.app.theme.colours.black
        row[HOSTNAME_COLUMN].setReadOnly(True)
        row[PORT_COLUMN].setText('Port')
        row[PORT_COLUMN].style.foreColour = self.app.theme.colours.black
        row[PORT_COLUMN].setReadOnly(True)

        self.serverTable.getRow(0).style.backColour = (
                self.app.theme.colours.grey)

    def setupCanvas(self):
        position = Location(ScaledScreenAttachedPoint(
                ScaledPoint(100, 200), 'topleft'), 'topleft')
        maxSize = 325 * self.app.screenManager.scaleFactor
        size = self.serverTable._getSize()
        if size[1] > maxSize:
            # We will be using the scrollbars
            displaySize = (size[0] + scrollableCanvas.ScrollBar.defaultWidth,
                    maxSize)
        else:
            displaySize = size
        self.canvas = scrollableCanvas.ScrollableCanvas(self.app, position,
                Size(*size), Size(*displaySize))
        self.serverTable.pos = Location(
                scrollableCanvas.ScrollableCanvasAttachedPoint(self.canvas,
                (0,0)), 'topleft')
        self.canvas.elements.append(self.serverTable)


    def addRow(self, host, port, httpAddress):
        index = len(self.httpAddresses) # 0-based index into list of servers
        row = self.serverTable.addRow()
        row[HOSTNAME_COLUMN].setText(host)
        row[HOSTNAME_COLUMN].onValueChanged.addListener(self.hostnameChanged)
        row[PORT_COLUMN].setText(str(port))
        row[PORT_COLUMN].setValidator(lambda text:(intValidator(text) and
                int(text) < 65536))
        row[PORT_COLUMN].setMaxLength(5)
        row[PORT_COLUMN].onValueChanged.addListener(self.portChanged)
        onClickConnect = lambda sender: self.join(index)
        onClickDelete = lambda sender: self.delete(index)
        row[CONNECT_COLUMN].setText('play')
        row[CONNECT_COLUMN].setOnClick(onClickConnect)

        onClickAccount = lambda sender: self.showAccountScreen(index)
        row[ACCOUNT_COLUMN].setText('account')
        row[ACCOUNT_COLUMN].setOnClick(onClickAccount)

        if index > 0:
            onClickUp = lambda sender: self.moveUp(index)
            row[MOVE_UP_COLUMN].setText('^')
            row[MOVE_UP_COLUMN].setOnClick(onClickUp)

            onClickDown = lambda sender: self.moveDown(index-1)
            self.serverTable.getRow(index)[MOVE_DOWN_COLUMN].setText('V')
            self.serverTable.getRow(index)[MOVE_DOWN_COLUMN].setOnClick(
                    onClickDown)
        else:
            row[MOVE_UP_COLUMN].setText('-')

        row[MOVE_DOWN_COLUMN].setText('-')
        row[DELETE_COLUMN].setText('X')
        row[DELETE_COLUMN].setOnClick(onClickDelete)
        self.httpAddresses.append(httpAddress)

    def moveUp(self, index):
        servers = list(self.app.connectionSettings.servers)
        temp = servers[index]
        servers[index] = servers[index-1]
        servers[index-1] = temp
        self.app.connectionSettings.servers = tuple(servers)
        self.app.connectionSettings.save()
        self.reload()

    def moveDown(self, index):
        servers = list(self.app.connectionSettings.servers)
        temp = servers[index]
        servers[index] = servers[index+1]
        servers[index+1] = temp
        self.app.connectionSettings.servers = tuple(servers)
        self.app.connectionSettings.save()
        self.reload()

    def delete(self, index):
        servers = list(self.app.connectionSettings.servers)
        del servers[index]
        self.app.connectionSettings.servers = tuple(servers)
        self.app.connectionSettings.save()
        self.reload()

    def newServer(self):
        servers = list(self.app.connectionSettings.servers)
        servers.append(('', 6787, ''))
        self.app.connectionSettings.servers = tuple(servers)
        self.app.connectionSettings.save()
        self.reload()

    def join(self, index):
        row = self.serverTable.getRow(index+1)
        hostname = row[HOSTNAME_COLUMN].getText()
        port = int(row[PORT_COLUMN].getText())
        if port > 65535:
            self.throwPortError()
        else:
            servers = [(hostname, port, '')]
            if self.app.connectionSettings.createGames:
                servers.append('create')
            self.onJoin(tuple(servers))

    def showAccountScreen(self, index):
        row = self.serverTable.getRow(index+1)
        hostname = row[HOSTNAME_COLUMN].getText()
        port = int(row[PORT_COLUMN].getText())
        if port > 65535:
            self.throwPortError()
        else:
            self.onAccountSettings(hostname, port)

    def joinLan(self):
        if self.app.connectionSettings.lanGames:
            self.onJoin(('lan',))
        else:
            self.onJoin(())

    def hostnameChanged(self, cell):
        self.httpAddresses[cell.rowNum - 1] = None
        self.save()

    def portChanged(self, cell):
        text = cell.getText()
        if text != '':
            port = int(text)
            if port > 65535:
                self.throwPortError()
                return
        # Else:
        self.save()

    def throwPortError(self):
        box = OkBox(self.app, ScaledSize(450, 150), 'Trosnoth',
                    'Port numbers must be less than 65536')
        box.show()

    def onJoinFailed(self):
        self.setElements()

    def setElements(self):
        self.elements = [self.canvas, self.lanButton, self.lanSearchButton,
                self.createButton, self.otherButton, self.newButton,
                self.closeButton]

    def save(self):
        servers = []
        for rowNum in range(0, len(self.httpAddresses)):
            row = self.serverTable.getRow(rowNum+1)
            hostname = row[HOSTNAME_COLUMN].getText()
            port = int(row[PORT_COLUMN].getText())
            if port < 1 or port > 65535:
                return
            httpAddress = self.httpAddresses[rowNum]
            if httpAddress is None: # Indicates a new or a changed row
                httpAddress = self.getDefaultHttpAddress(hostname, port)
            servers.append((hostname, port, httpAddress))

        self.app.connectionSettings.createGames = self.createButton.getValue()
        self.app.connectionSettings.lanGames = ('afterinet' if
                self.lanButton.getValue() else 'never')
        self.app.connectionSettings.otherGames = self.otherButton.getValue()
        self.app.connectionSettings.servers = tuple(servers)
        self.app.connectionSettings.save()

    def getDefaultHttpAddress(self, hostname, port):
        if (hostname, port) == ('localhost', 6787):
            return 'http://localhost:8080/'
        else:
            return 'http://%s/' % (hostname,)

    def reload(self):
        self.httpAddresses = []
        self.setupTable()
        for server in self.app.connectionSettings.servers:
            self.addRow(*server)
        self.setupCanvas()
        self.setElements()

class ChangePasswordTab(Tab, framework.TabFriendlyCompoundElement):
    def __init__(self, app, host, onClose, onSave):
        super(ChangePasswordTab, self).__init__(app, 'Password')
        self.onClose = onClose
        self.onSave = onSave
        self.host = host

        font = app.screenManager.fonts.defaultTextBoxFont
        labelColour = app.theme.colours.dialogBoxTextColour
        inputColour = app.theme.colours.white

        label = TextElement(self.app,
            'Use this form to change your password for %s' % (self.host,),
            font=self.app.screenManager.fonts.smallMenuFont,
            pos=Location(ScaledScreenAttachedPoint(
                ScaledPoint(100, 200), 'topleft'), 'topleft'),
            colour=self.app.theme.colours.mainMenuColour,
        )

        self.passwordField = PasswordBox(app,
            Region(topleft=self.Relative(0.1, 0.48),
                bottomright=self.Relative(0.9, 0.64)),
                font=font, colour=inputColour,
                onClick=self.setFocus, onTab=self.tabNext)

        self.passwordField2 = PasswordBox(app,
            Region(topleft=self.Relative(0.1, 0.8),
                bottomright=self.Relative(0.9, 0.96)),
                font=font, colour=inputColour,
                onClick=self.setFocus, onTab=self.tabNext)

        self.elements = [
            label,
            TextElement(app, 'New password', font,
                Location(self.Relative(0.1, 0.43), 'midleft'),
                labelColour),
            self.passwordField,
            TextElement(app, 'Retype new password', font,
                Location(self.Relative(0.1, 0.75), 'midleft'),
                labelColour),
            self.passwordField2,
            button(app, 'save', self.onSave, (-100, -75), 'midbottom',
                    secondColour=app.theme.colours.white),
            button(app, 'cancel', self.cancel, (100, -75), 'midbottom',
                    secondColour=app.theme.colours.white),
        ]

        self.tabOrder = [self.passwordField, self.passwordField2]

    @defer.inlineCallbacks
    def save(self, protocol):
        try:
            if self.passwordField.value != self.passwordField2.value:
                defer.returnValue('Passwords do not match')
                return
            if self.passwordField.value:
                data = yield encryptPassword(protocol, self.passwordField.value)
                yield protocol.callRemote(authcommands.SetPassword,
                        password=data)
        except Exception:   # Must not be naked or returnValue is caught
            log.error('Error storing password', exc_info=True)
            defer.returnValue('Error saving settings')
            return

        defer.returnValue('')
        return

    def cancel(self):
        self.onClose()



from trosnoth.utils.event import Event
from trosnoth.gui.common import (Region, Canvas, Location)
from trosnoth.gui.framework import framework
from trosnoth.gui.framework.elements import (SolidRect, TextButton)
from trosnoth.model import mapLayout
from trosnoth.game import LocalGame

SIZE = (3, 1)
AICOUNT = 5

class PractiseScreen(framework.CompoundElement):
    def __init__(self, app, onClose=None, onStart=None):
        super(PractiseScreen, self).__init__(app)

        self.onClose = Event()
        if onClose is not None:
            self.onClose.addListener(onClose)
        self.onStart = Event()
        if onStart is not None:
            self.onStart.addListener(onStart)

        if app.displaySettings.alphaOverlays:
            alpha = 192
        else:
            alpha = None
        bg = SolidRect(app, app.theme.colours.playMenu, alpha,
                Region(centre=Canvas(512, 384), size=Canvas(924, 500)))

        #colour = app.theme.colours.mainMenuColour
        #font = app.screenManager.fonts.consoleFont

        font = app.screenManager.fonts.bigMenuFont
        cancel = TextButton(app, Location(Canvas(512, 624), 'midbottom'),
                'Cancel', font, app.theme.colours.secondMenuColour,
                app.theme.colours.white, onClick=self.cancel)
        self.elements = [bg, cancel]

    def cancel(self, element):
        self.onClose.execute()

    def startGame(self):
        db = self.app.layoutDatabase
        game = LocalGame(db, SIZE[0], SIZE[1],
                onceOnly=True)
        game.world.setNextGameType('solo')
        game.world.beginGameStartCountdown()

        for i in xrange(AICOUNT):
            game.addAI('ranger')

        self.onStart.execute(game)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import framework, elements, hotkey, tabContainer, tab, statList, prompt
from trosnoth.gui.fonts.font import Font
from trosnoth.gui.common import *
import pygame

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        bg = pygame.surface.Surface((1024,768))
        bg.fill((255,255,0))
                                                                     
        font = Font("KLEPTOCR.TTF", 30)
        tc = tabContainer.TabContainer(app, ScaledArea(10,10, 1004, 748), font, (0,128,0))
        background = SizedImage('Awesome Leaders.jpg', tabContainer.TabSize(tc))
        for x in range(0,1):
            t1 = tab.Tab(app, "Hellooooooooo")
            t1.elements = [elements.PictureElement(app, background, Location(AttachedPoint((0,0),tc._getTabRect,'topleft'),'topleft'))]
            t2 = NewTab(app, tc)
            t3 = EpilepticTab(app, tc)
            tc.addTab(t1)
            tc.addTab(t2)
            tc.addTab(t3)
        #tc._setTabHeaderDimensions(3,36,24,200)
        self.elements = [elements.Backdrop(app, bg), tc]

class NewTab(tab.Tab):
    def __init__(self, app, tabContainer):
        super(NewTab, self).__init__(app, "New Tab")
        h = prompt.InputBox(app, Area(AttachedPoint((10,10), tabContainer._getTabRect), ScaledSize(400,100)), "Press Enter", font = app.screenManager.fonts.bigMenuFont)
        h.onEnter.addListener(lambda sender:self.app.screenManager.setScreenProperties((1024,768), 0, "Testing"))
        h.onClick.addListener(self.setFocus)
        self.elements = [h]

import pygame
from random import randint, shuffle


class EpilepticTab(tab.Tab):
    def __init__(self, app, tabContainer):
        super(EpilepticTab, self).__init__(app, str(randint(0,1000)))
        self.val = 4
        self.l = ['hello', self.val]
        getColour = lambda item: ((255,0,0), (0,255,255))['4' in str(item)]
        font = Font("KLEPTOCR.TTF", 50)
        self.s = statList.StatList(app,
                              Location(AttachedPoint((10,100), tabContainer._getTabRect)),
                              self.l,
                              font,
                              colourFunction = getColour)
        self.s.setVisible()
        addTabFunction = lambda sender: tabContainer.addTab(EpilepticTab(app, tabContainer))
        closeTabFunction = lambda sender: tabContainer.removeTab(self)
        button = elements.TextButton(app, Location(AttachedPoint((200,50), tabContainer._getTabRect)), "Click for new Tab", font, (0,128,0), (0,255,0))
        button.onClick.addListener(addTabFunction)
        otherButton = elements.TextButton(app, Location(AttachedPoint((200,90), tabContainer._getTabRect)), "Close This Tab", font, (128,0,0), (255,0,0))
        otherButton.onClick.addListener(closeTabFunction)

        f = lambda: self.s.setVisible(not self.s.getVisibility())

        h = hotkey.Hotkey(app, pygame.K_s, 0)
        h.onTriggered.addListener(f)
        # TODO: add a backdrop
        self.elements += [self.s, button, otherButton, h]
    def tick(self, deltaT):
        super(EpilepticTab, self).tick(deltaT)
        if randint(0,100) == 1:
            shuffle(self.l)
        if randint(0,30000) == 42:
            self.l.append("What's going on here?! %d" % randint(0,100000))
        if randint(0,1000) == 100:
            i = self.l.index(self.val)
            self.val = randint(0, 1000)
            self.l[i] = self.val
size = (600,450)
if __name__ == '__main__':
    a = Application(size, 0, "Testing", Interface)
    a.run()



import datetime
import logging

from trosnoth.gui.framework import framework
from trosnoth.utils.event import Event

log = logging.getLogger(__name__)


class ITimer(framework.Element):
    def __init__(self, clock=None):
        self.running = False
        self.clock = clock
        if clock:
            self.lastClockTime = clock()
        else:
            self.lastClockTime = None

    def getDelta(self, delta):
        if self.clock:
            now = self.clock()
            delta = now - self.lastClockTime
            self.lastClockTime = now
        return delta

    def start(self):
        self.running = True
        return self

    def pause(self):
        self.running = False
        return self

    def tick(self, deltaT):
        raise NotImplementedError('%s.tick' % (self.__class__.__name__,))

    def getTimeString(self):
        raise NotImplementedError(
            '%s.getTimeString' % (self.__class__.__name__,))


class CountdownTimer(ITimer):
    '''
    @param amount    Amount of time to count down
    '''

    def __init__(self, countTo=0, highest="days", *args, **kwargs):
        super(CountdownTimer, self).__init__(*args, **kwargs)
        self.countTo = countTo
        self.counted = 0
        self.onCountedDown = Event()
        self.highest = highest

    def tick(self, deltaT):
        deltaT = self.getDelta(deltaT)
        if self.running:
            self.counted += deltaT

            if self.counted > self.countTo:
                self.counted = self.countTo
                self.onCountedDown.execute()

    def getCurTime(self):
        return max(0, self.countTo - self.counted)

    def getTimeString(self):
        total = int(self.getCurTime())
        deltaResult = datetime.timedelta(seconds=total)
        seconds = deltaResult.days * 24 * 60 * 60 + deltaResult.seconds
        if self.highest == "seconds":
            return str(seconds)
        elif self.highest == "minutes":
            minutes, seconds = divmod(seconds, 60)
            return "%02d:%02d" % (minutes, seconds)
        elif self.highest == "hours":
            hours, minutes = divmod(minutes, 60)
            return "%d:%02d:%02d" % (hours, minutes, seconds)
        else:
            return str(deltaResult)


class Timer(ITimer):
    '''@param tick  Frequency with which onTick events are triggered.'''
    def __init__(
            self, tickDuration=1, startAt=None, highest="days",
            *args, **kwargs):
        super(Timer, self).__init__(*args, **kwargs)
        if startAt is None:
            self.counted = 0
        else:
            self.counted = startAt
        self.tickDuration = tickDuration
        self.highest = highest

        self.onTick = Event()

    def tick(self, deltaT):
        deltaT = self.getDelta(deltaT)
        if self.running:
            numTicked = self.counted / self.tickDuration
            self.counted += deltaT
            if (self.counted / self.tickDuration) > numTicked:
                self.onTick.execute()

    def getTimeString(self):
        deltaResult = datetime.timedelta(seconds=int(max(0, self.counted)))
        seconds = deltaResult.days * 24 * 60 * 60 + deltaResult.seconds
        if self.highest == "seconds":
            return str(seconds)
        elif self.highest == "minutes":
            minutes, seconds = divmod(seconds, 60)
            return "%02d:%02d" % (minutes, seconds)
        elif self.highest == "hours":
            hours, minutes = divmod(minutes, 60)
            return "%d:%02d:%02d" % (hours, minutes, seconds)
        else:
            return str(deltaResult)
